from __future__ import annotations

import math
import os
import random
import re
from collections.abc import Generator, Sequence
from typing import TYPE_CHECKING, Any

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import torch
from omegaconf import DictConfig
from sklearn.metrics import ConfusionMatrixDisplay, accuracy_score, classification_report, confusion_matrix
from sklearn.model_selection import StratifiedKFold, StratifiedShuffleSplit
from torch.utils.data import DataLoader

from quadra.models.base import ModelSignatureWrapper
from quadra.utils import utils
from quadra.utils.models import get_feature
from quadra.utils.visualization import UnNormalize, plot_classification_results

if TYPE_CHECKING:
    from quadra.datamodules.classification import SklearnClassificationDataModule
    from quadra.datamodules.patch import PatchSklearnClassificationDataModule

log = utils.get_logger(__name__)


def get_file_condition(
    file_name: str, root: str, exclude_filter: list[str] | None = None, include_filter: list[str] | None = None
):
    """Check if a file should be included or excluded based on the filters provided.

    Args:
        file_name: Name of the file
        root: Root directory of the file
        exclude_filter: List of string filter to be used to exclude images. If None no filter will be applied.
        include_filter: List of string filter to be used to include images. If None no filter will be applied.
    """
    if exclude_filter is not None:
        if any(fil in file_name for fil in exclude_filter):
            return False

        if any(fil in root for fil in exclude_filter):
            return False

    return not (
        include_filter is not None
        and (not any(fil in file_name for fil in include_filter) and not any(fil in root for fil in include_filter))
    )


def natural_key(string_):
    """See http://www.codinghorror.com/blog/archives/001018.html."""
    return [int(s) if s.isdigit() else s for s in re.split(r"(\d+)", string_.lower())]


def find_images_and_targets(
    folder: str,
    types: list | None = None,
    class_to_idx: dict[str, int] | None = None,
    leaf_name_only: bool = True,
    sort: bool = True,
    exclude_filter: list | None = None,
    include_filter: list | None = None,
    label_map: dict[str, Any] | None = None,
) -> tuple[np.ndarray, np.ndarray, dict]:
    """Given a folder, extract the absolute path of all the files with a valid extension.
    Then assign a label based on subfolder name.

    Args:
        folder: path to main folder
        types: valid file extentions
        class_to_idx: dictionary of conversion btw folder name and index.
            Only file whose label is in dictionary key list will be considered. If None all files will
            be considered and a custom conversion is created.
        leaf_name_only: if True use only the leaf folder name as label, otherwise use the full path
        sort: if True sort the images and labels based on the image name
        exclude_filter: list of string filter to be used to exclude images.
            If None no filter will be applied.
        include_filter: list of string filder to be used to include images.
            Only images that satisfied at list one of the filter will be included.
        label_map: dictionary of conversion btw folder name and label.
    """
    if types is None:
        types = [".png", ".jpg", ".jpeg", ".bmp"]
    labels = []
    filenames = []

    for root, _, files in os.walk(folder, topdown=False, followlinks=True):
        if root != folder:
            rel_path = os.path.relpath(root, folder)
        else:
            rel_path = ""

        if leaf_name_only:
            label = os.path.basename(rel_path)
        else:
            aa = rel_path.split(os.path.sep)
            if len(aa) == 2:
                aa = aa[-1:]
            else:
                aa = aa[-2:]
            label = "_".join(aa)  # rel_path.replace(os.path.sep, "_")
            # label = rel_path.replace(os.path.sep, "_")

        for f in files:
            if not get_file_condition(
                file_name=f, root=root, exclude_filter=exclude_filter, include_filter=include_filter
            ):
                continue

            if f.startswith(".") or "checkpoint" in f:
                continue
            _, ext = os.path.splitext(f)
            if ext.lower() in types:
                filenames.append(os.path.join(root, f))
                labels.append(label)

    if label_map is not None:
        labels, _ = group_labels(labels, label_map)

    if class_to_idx is None:
        # building class index
        unique_labels = set(labels)
        sorted_labels = sorted(unique_labels, key=natural_key)
        class_to_idx = {str(c): idx for idx, c in enumerate(sorted_labels)}

    images_and_targets = [(f, l) for f, l in zip(filenames, labels, strict=False) if l in class_to_idx]

    if sort:
        images_and_targets = sorted(images_and_targets, key=lambda k: natural_key(k[0]))

    return np.array(images_and_targets)[:, 0], np.array(images_and_targets)[:, 1], class_to_idx


def find_test_image(
    folder: str,
    types: list[str] | None = None,
    exclude_filter: list[str] | None = None,
    include_filter: list[str] | None = None,
    include_none_class: bool = True,
    test_split_file: str | None = None,
    label_map=None,
) -> tuple[list[str], list[str | None]]:
    """Given a path extract images and labels with filters, labels are based on the parent folder name of the images
    Args:
        folder: root directory containing the images
        types: only choose images with the extensions specified, if None use default extensions
        exclude_filter: list of string filter to be used to exclude images. If None no filter will be applied.
        include_filter: list of string filter to be used to include images. If None no filter will be applied.
        include_none_class: if set to True convert all 'None' labels to None, otherwise ignore the image
        test_split_file: if defined use the split defined inside the file
    Returns:
        Two lists, one containing the images path and the other one containing the labels. Labels can be None.
    """
    if types is None:
        types = [".png", ".jpg", ".jpeg", ".bmp"]

    labels = []
    filenames = []

    for root, _, files in os.walk(folder, topdown=False, followlinks=True):
        rel_path = os.path.relpath(root, folder) if root != folder else ""
        label: str | None = os.path.basename(rel_path)
        for f in files:
            if not get_file_condition(
                file_name=f, root=root, exclude_filter=exclude_filter, include_filter=include_filter
            ):
                continue
            if f.startswith(".") or "checkpoint" in f:
                continue
            _, ext = os.path.splitext(f)
            if ext.lower() in types:
                if label == "None":
                    if include_none_class:
                        label = None
                    else:
                        continue
                filenames.append(os.path.join(root, f))
                labels.append(label)

    if test_split_file is not None:
        if not os.path.isabs(test_split_file):
            log.info(
                "test_split_file is not an absolute path. Trying to using folder argument %s as parent folder", folder
            )
            test_split_file = os.path.join(folder, test_split_file)

        if not os.path.exists(test_split_file):
            raise FileNotFoundError(f"test_split_file {test_split_file} does not exist")

        with open(test_split_file) as test_file:
            test_split = test_file.read().splitlines()

        file_samples = []
        for row in test_split:
            csv_values = row.split(",")
            if len(csv_values) == 1:
                # ensuring backward compatibility with old split file format
                # old_format: sample, new_format: sample,class
                sample_path = os.path.join(folder, csv_values[0])
            else:
                sample_path = os.path.join(folder, ",".join(csv_values[:-1]))

            file_samples.append(sample_path)

        test_split = [os.path.join(folder, sample.strip()) for sample in file_samples]
        labels = [t for s, t in zip(filenames, labels, strict=False) if s in file_samples]
        filenames = [s for s in filenames if s in file_samples]
        log.info("Selected %d images using test_split_file for the test", len(filenames))
        if len(filenames) != len(file_samples):
            log.warning(
                "test_split_file contains %d images but only %d images were found in the folder."
                "This may be due to duplicate lines in the test_split_file.",
                len(file_samples),
                len(filenames),
            )
    else:
        log.info("No test_split_file. Selected all %s images for the test", folder)

    if label_map is not None:
        labels, _ = group_labels(labels, label_map)

    return filenames, labels


def group_labels(labels: Sequence[str | None], class_mapping: dict[str, str | None | list[str]]) -> tuple[list, dict]:
    """Group labels based on class_mapping.

    Raises:
        ValueError: if a label is not in class_mapping
        ValueError: if a label is in class_mapping but has no corresponding value

    Returns:
       List of labels and a dictionary of labels and their corresponding group

    Example:
        ```python
        grouped_labels, class_to_idx = group_labels(labels, class_mapping={"Good": "A", "Bad": None})
        assert grouped_labels.count("Good") == labels.count("A")
        assert len(class_to_idx.keys()) == 2

        grouped_labels, class_to_idx = group_labels(labels, class_mapping={"Good": "A", "Defect": "B", "Bad": None})
        assert grouped_labels.count("Bad") == labels.count("C") + labels.count("D")
        assert len(class_to_idx.keys()) == 3

        grouped_labels, class_to_idx = group_labels(labels, class_mapping={"Good": "A", "Bad": ["B", "C", "D"]})
        assert grouped_labels.count("Bad") == labels.count("B") + labels.count("C") + labels.count("D")
        assert len(class_to_idx.keys()) == 2
        ```
    """
    grouped_labels = []
    specified_targets = [k for k in class_mapping if class_mapping[k] is not None]
    non_specified_targets = [k for k in class_mapping if class_mapping[k] is None]
    if len(non_specified_targets) > 1:
        raise ValueError(f"More than one non specified target: {non_specified_targets}")
    for label in labels:
        found = False
        for target in specified_targets:
            if not found:
                current_mapping = class_mapping[target]
                if current_mapping is None:
                    continue

                if any(label in list(related_label) for related_label in current_mapping if related_label is not None):
                    grouped_labels.append(target)
                    found = True
        if not found:
            if len(non_specified_targets) > 0:
                grouped_labels.append(non_specified_targets[0])
            else:
                raise ValueError(f"No target found for label: {label}")
    class_to_idx = {k: i for i, k in enumerate(class_mapping.keys())}
    return grouped_labels, class_to_idx


def filter_with_file(list_of_full_paths: list[str], file_path: str, root_path: str) -> tuple[list[str], list[bool]]:
    """Filter a list of items using a file containing the items to keep. Paths inside file
    should be relative to root_path not absolute to avoid user related issues.

    Args:
        list_of_full_paths: list of items to filter
        file_path: path to the file containing the items to keep
        root_path: root path of the dataset

    Returns:
        list of items to keep
        the mask list to apply different lists later.
    """
    filtered_full_paths = []
    filter_mask = []

    with open(file_path) as f:
        for relative_path in f.read().splitlines():
            full_path = os.path.join(root_path, relative_path)
            if full_path in list_of_full_paths:
                filtered_full_paths.append(full_path)
                filter_mask.append(True)
            else:
                filter_mask.append(False)

    return filtered_full_paths, filter_mask


def get_split(
    image_dir: str,
    exclude_filter: list[str] | None = None,
    include_filter: list[str] | None = None,
    test_size: float = 0.3,
    random_state: int = 42,
    class_to_idx: dict[str, int] | None = None,
    label_map: dict | None = None,
    n_splits: int = 1,
    include_none_class: bool = False,
    limit_training_data: int | None = None,
    train_split_file: str | None = None,
) -> tuple[np.ndarray, np.ndarray, Generator[list, None, None], dict]:
    """Given a folder, extract the absolute path of all the files with a valid extension and name
    and split them into train/test.

    Args:
        image_dir: Path to the folder containing the images
        exclude_filter: List of file name filter to be excluded: If None no filter will be applied
        include_filter: List of file name filter to be included: If None no filter will be applied
        test_size: Percentage of data to be used for test
        random_state: Random state to be used for reproducibility
        class_to_idx: Dictionary of conversion btw folder name and index.
            Only file whose label is in dictionary key list will be considered.
            If None all files will be considered and a custom conversion is created.
        label_map: Dictionary of conversion btw folder name and label.
        n_splits: Number of dataset subdivision (default 1 -> train/test)
        include_none_class: If set to True convert all 'None' labels to None
        limit_training_data: If set to a value, limit the number of training samples to this value
        train_split_file: If set to a path, use the file to split the dataset
    """
    # TODO: Why is include_none_class not used?
    # pylint: disable=unused-argument
    assert os.path.isdir(image_dir), f"Folder {image_dir} does not exist."
    # Get samples and target
    samples, targets, class_to_idx = find_images_and_targets(
        folder=image_dir,
        exclude_filter=exclude_filter,
        include_filter=include_filter,
        class_to_idx=class_to_idx,
        label_map=label_map,
        # include_none_class=include_none_class,
    )

    cl, counts = np.unique(targets, return_counts=True)

    for num, _cl in zip(counts, cl, strict=False):
        if num == 1:
            to_remove = np.where(np.array(targets) == _cl)[0][0]
            samples = np.delete(np.array(samples), to_remove)
            targets = np.delete(np.array(targets), to_remove)
            class_to_idx.pop(_cl)

    if train_split_file is not None:
        with open(train_split_file) as f:
            train_split = f.read().splitlines()

        file_samples = []
        for row in train_split:
            csv_values = row.split(",")

            if len(csv_values) == 1:
                # ensuring backward compatibility with the old split file format
                # old_format: sample, new_format: sample,class
                sample_path = os.path.join(image_dir, csv_values[0])
            else:
                sample_path = os.path.join(image_dir, ",".join(csv_values[:-1]))

            file_samples.append(sample_path)

        train_split = [os.path.join(image_dir, sample.strip()) for sample in file_samples]
        targets = np.array([t for s, t in zip(samples, targets, strict=False) if s in file_samples])
        samples = np.array([s for s in samples if s in file_samples])

    if limit_training_data is not None:
        idx_to_keep = []
        for cl in np.unique(targets):
            cl_idx = np.where(np.array(targets) == cl)[0].tolist()
            random.seed(random_state)
            random.shuffle(cl_idx)
            idx_to_keep.extend(cl_idx[:limit_training_data])

        samples = np.asarray([samples[i] for i in idx_to_keep])
        targets = np.asarray([targets[i] for i in idx_to_keep])

    _, counts = np.unique(targets, return_counts=True)

    if n_splits == 1:
        split_technique = StratifiedShuffleSplit(n_splits=1, test_size=test_size, random_state=random_state)
    else:
        split_technique = StratifiedKFold(n_splits=n_splits, random_state=random_state, shuffle=True)

    split = split_technique.split(samples, targets)

    return np.array(samples), np.array(targets), split, class_to_idx


def save_classification_result(
    results: pd.DataFrame,
    output_folder: str,
    test_dataloader: DataLoader,
    config: DictConfig,
    output: DictConfig,
    accuracy: float | None = None,
    confmat: pd.DataFrame | None = None,
    grayscale_cams: np.ndarray | None = None,
):
    """Save csv results, confusion matrix and example images.

    Args:
        results: Dataframe containing the results
        output_folder: Path to the output folder
        confmat: Confusion matrix in a pandas dataframe, may be None if all test labels are unknown
        accuracy: Accuracy of the model, is None if all test labels are unknown
        test_dataloader: Dataloader used for testing
        config: Configuration file
        output: Output configuration
        grayscale_cams: List of grayscale grad_cam outputs ordered as the results
    """
    # Save csv
    results.to_csv(os.path.join(output_folder, "test_results.csv"), index_label="index")
    if grayscale_cams is None:
        log.info("Plotting only original examples, set gradcam = true in config file to also plot gradcam examples")

        save_gradcams = False
    else:
        log.info("Plotting original and gradcam examples")
        save_gradcams = True

    if confmat is not None and accuracy is not None:
        # Save confusion matrix
        disp = ConfusionMatrixDisplay(
            confusion_matrix=np.array(confmat),
            display_labels=[x.replace("pred:", "") for x in confmat.columns.to_list()],
        )
        disp.plot(include_values=True, cmap=plt.cm.Greens, ax=None, colorbar=False, xticks_rotation=90)
        plt.title(f"Confusion Matrix (Accuracy: {(accuracy * 100):.2f}%)")
        plt.savefig(
            os.path.join(output_folder, "test_confusion_matrix.png"),
            bbox_inches="tight",
            pad_inches=0,
            dpi=300,
        )
        plt.close()

    if output is not None and output.example:
        log.info("Saving discordant/concordant examples in test folder")
        idx_to_class = test_dataloader.dataset.idx_to_class  # type: ignore[attr-defined]

        # Get misclassified samples
        images_folder = os.path.join(output_folder, "example")
        if not os.path.isdir(images_folder):
            os.makedirs(images_folder)
        original_images_folder = os.path.join(images_folder, "original")
        if not os.path.isdir(original_images_folder):
            os.makedirs(original_images_folder)

        gradcam_folder = os.path.join(images_folder, "gradcam")
        if save_gradcams and not os.path.isdir(gradcam_folder):
            os.makedirs(gradcam_folder)

        for v in np.unique([results["real_label"], results["pred_label"]]):
            if np.isnan(v) or v == -1:
                continue

            k = idx_to_class[v]
            plot_classification_results(
                test_dataloader.dataset,
                unorm=UnNormalize(mean=config.transforms.mean, std=config.transforms.std),
                pred_labels=results["pred_label"].to_numpy(),
                test_labels=results["real_label"].to_numpy(),
                grayscale_cams=grayscale_cams,
                class_name=k,
                original_folder=original_images_folder,
                gradcam_folder=gradcam_folder,
                idx_to_class=idx_to_class,
                pred_class_to_plot=v,
                what="con",
                rows=output.get("rows", 3),
                cols=output.get("cols", 2),
                figsize=output.get("figsize", (20, 20)),
                gradcam=save_gradcams,
            )

            plot_classification_results(
                test_dataloader.dataset,
                unorm=UnNormalize(mean=config.transforms.mean, std=config.transforms.std),
                pred_labels=results["pred_label"].to_numpy(),
                test_labels=results["real_label"].to_numpy(),
                grayscale_cams=grayscale_cams,
                class_name=k,
                original_folder=original_images_folder,
                gradcam_folder=gradcam_folder,
                idx_to_class=idx_to_class,
                pred_class_to_plot=v,
                what="dis",
                rows=output.get("rows", 3),
                cols=output.get("cols", 2),
                figsize=output.get("figsize", (20, 20)),
                gradcam=save_gradcams,
            )

    else:
        log.info("Not generating discordant/concordant examples. Check task.output.example in config file")


def get_results(
    test_labels: np.ndarray | list[int],
    pred_labels: np.ndarray | list[int],
    idx_to_labels: dict | None = None,
    cl_rep_digits: int = 3,
) -> tuple[str | dict, pd.DataFrame, float]:
    """Get prediction results from predicted and test labels.

    Args:
        test_labels : test labels
        pred_labels : predicted labels
        idx_to_labels : dictionary mapping indices to labels
        cl_rep_digits : number of digits to use in the classification report. Default: 3

    Returns:
        A tuple that contains classification report as dictionary, `cm` is a pd.Dataframe representing
        the Confusion Matrix, acc is the computed accuracy
    """
    unique_labels = np.unique([test_labels, pred_labels])
    cl_rep = classification_report(
        y_true=test_labels,
        y_pred=pred_labels,
        labels=unique_labels,
        digits=cl_rep_digits,
        zero_division=0,
    )

    cm = confusion_matrix(y_true=test_labels, y_pred=pred_labels, labels=unique_labels)

    acc = accuracy_score(y_true=test_labels, y_pred=pred_labels)

    if idx_to_labels:
        pd_cm = pd.DataFrame(
            cm,
            index=[f"true:{idx_to_labels[x]}" for x in unique_labels],
            columns=[f"pred:{idx_to_labels[x]}" for x in unique_labels],
        )
    else:
        pd_cm = pd.DataFrame(
            cm,
            index=[f"true:{x}" for x in unique_labels],
            columns=[f"pred:{x}" for x in unique_labels],
        )
    return cl_rep, pd_cm, acc


def automatic_batch_size_computation(
    datamodule: SklearnClassificationDataModule | PatchSklearnClassificationDataModule,
    backbone: ModelSignatureWrapper,
    starting_batch_size: int,
) -> int:
    """Find the optimal batch size for feature extraction. This algorithm works from the largest batch size possible
    and divide by 2 until it finds the largest batch size that fits in memory.

    Args:
        datamodule: Datamodule used for feature extraction
        backbone: Backbone used for feature extraction
        starting_batch_size: Starting batch size to use for the search

    Returns:
        Optimal batch size
    """
    log.info("Finding optimal batch size...")
    optimal = False
    batch_size = starting_batch_size

    while not optimal:
        datamodule.batch_size = batch_size
        base_dataloader = datamodule.train_dataloader()

        if isinstance(base_dataloader, Sequence):
            base_dataloader = base_dataloader[0]

        if len(base_dataloader) == 1:
            # If it fits in memory this is the largest batch size possible
            # If it crashes restart with the previous batch size // 2
            datamodule.batch_size = len(base_dataloader.dataset)  # type: ignore[arg-type]
            # New restarting batch size is the largest closest power of 2 to the dataset size, it will be divided by 2
            batch_size = 2 ** math.ceil(math.log2(datamodule.batch_size))
            base_dataloader = datamodule.train_dataloader()
            if isinstance(base_dataloader, Sequence):
                base_dataloader = base_dataloader[0]
            optimal = True

        try:
            log.info("Trying batch size: %d", datamodule.batch_size)
            _ = get_feature(feature_extractor=backbone, dl=base_dataloader, iteration_over_training=1, limit_batches=1)
        except RuntimeError as e:
            if batch_size > 1:
                batch_size = batch_size // 2
                optimal = False
                continue

            log.error("Unable to run the model with batch size 1")
            raise e

        log.info("Found optimal batch size: %d", datamodule.batch_size)
        optimal = True

    if torch.cuda.is_available():
        torch.cuda.empty_cache()

    return datamodule.batch_size



from __future__ import annotations

import glob
import itertools
import json
import math
import os
import random
import shutil
import warnings
from collections.abc import Callable
from copy import deepcopy
from dataclasses import dataclass
from functools import partial
from multiprocessing import Pool
from typing import Any

import cv2
import h5py
import numpy as np
from scipy import ndimage
from skimage.measure import label, regionprops  # pylint: disable=no-name-in-module
from skimage.util import view_as_windows
from skmultilearn.model_selection import iterative_train_test_split
from tqdm import tqdm
from tripy import earclip

from quadra.utils import utils

log = utils.get_logger(__name__)


@dataclass
class PatchDatasetFileFormat:
    """Model representing the content of the patch dataset split_files field in the info.json file."""

    image_path: str
    mask_path: str | None = None


@dataclass
class PatchDatasetInfo:
    """Model representing the content of the patch dataset info.json file."""

    patch_size: tuple[int, int] | None
    patch_number: tuple[int, int] | None
    annotated_good: list[int] | None
    overlap: float
    train_files: list[PatchDatasetFileFormat]
    val_files: list[PatchDatasetFileFormat]
    test_files: list[PatchDatasetFileFormat]

    @staticmethod
    def _map_files(files: list[Any]):
        """Convert a list of dict to a list of PatchDatasetFileFormat."""
        mapped_files = []
        for file in files:
            current_file = file
            if isinstance(file, dict):
                current_file = PatchDatasetFileFormat(**current_file)
            mapped_files.append(current_file)

        return mapped_files

    def __post_init__(self):
        self.train_files = self._map_files(self.train_files)
        self.val_files = self._map_files(self.val_files)
        self.test_files = self._map_files(self.test_files)


def get_image_mask_association(
    data_folder: str,
    mask_folder: str | None = None,
    mask_extension: str = "",
    warning_on_missing_mask: bool = True,
) -> list[dict]:
    """Function used to match images and mask from a folder or sub-folders.

    Args:
        data_folder: root data folder containing images or images and masks
        mask_folder: Optional root directory used to search only the masks
        mask_extension: extension used to identify the mask file, it's mandatory if mask_folder is not specified
            warning_on_missing_mask: if set to True a warning will be raised if a mask is missing, disable if you know
            that many images do not have a mask.
        warning_on_missing_mask: if set to True a warning will be raised if a mask is missing, disable if you know

    Returns:
        List of dict like:
        [
        {
            'base_name': '161927.tiff',
            'path': 'test_dataset_patch/images/161927.tiff',
            'mask': 'test_dataset_patch/masks/161927_mask.tiff'
        }, ...
        ]
    """
    # get all the images from the data folder
    data_images = glob.glob(os.path.join(data_folder, "**", "*"), recursive=True)

    basenames = [os.path.splitext(os.path.basename(image))[0] for image in data_images]

    if len(set(basenames)) != len(basenames):
        raise ValueError("Found multiple images with the same name and different extension, this is not supported.")

    log.info("Found: %d images in %s", len(data_images), data_folder)
    # divide images and mask if in the same folder
    # if mask folder is specified search mask in that folder
    if mask_folder:
        masks_images = []
        for basename in basenames:
            mask_path = os.path.join(mask_folder, f"{basename}{mask_extension}.*")
            mask_path_list = glob.glob(mask_path)

            if len(mask_path_list) == 1:
                masks_images.append(mask_path_list[0])
            elif warning_on_missing_mask:
                log.warning("Mask for %s not found", basename)
    else:
        if mask_extension == "":
            raise ValueError("If no mask folder is provided, mask extension is mandatory it cannot be empty.")

        masks_images = [image for image in data_images if mask_extension in image]
        data_images = [image for image in data_images if mask_extension not in image]

    # build support dictionary
    unique_images = [{"base_name": os.path.basename(image), "path": image, "mask": None} for image in data_images]

    images_stem = [os.path.splitext(str(image["base_name"]))[0] + mask_extension for image in unique_images]
    masks_stem = [os.path.splitext(os.path.basename(mask))[0] for mask in masks_images]

    # search corrispondency between file or folders
    for i, image_stem in enumerate(images_stem):
        if image_stem in masks_stem:
            unique_images[i]["mask"] = masks_images[masks_stem.index(image_stem)]

    log.info("Unique images with mask: %d", len([uni for uni in unique_images if uni.get("mask") is not None]))
    log.info("Unique images with no mask: %d", len([uni for uni in unique_images if uni.get("mask") is None]))

    return unique_images


def compute_patch_info(
    img_h: int,
    img_w: int,
    patch_num_h: int,
    patch_num_w: int,
    overlap: float = 0.0,
) -> tuple[tuple[int, int], tuple[int, int]]:
    """Compute the patch size and step size given the number of patches and the overlap.

    Args:
        img_h: height of the image
        img_w: width of the image
        patch_num_h: number of vertical patches
        patch_num_w: number of horizontal patches
        overlap: percentage of overlap between patches.

    Returns:
        Tuple containing:
            patch_size: [size_y, size_x] Dimension of the patch
            step_size: [step_y, step_x]  Step size
    """
    patch_size_h = np.ceil(img_h / (1 + (patch_num_h - 1) - (patch_num_h - 1) * overlap)).astype(int)
    step_h = patch_size_h - np.ceil(overlap * patch_size_h).astype(int)

    patch_size_w = np.ceil(img_w / (1 + (patch_num_w - 1) - (patch_num_w - 1) * overlap)).astype(int)
    step_w = patch_size_w - np.ceil(overlap * patch_size_w).astype(int)

    # We want a combination of patch size and step that if the image is not divisible by the number of patches
    # will try to fit the maximum number of patches in the image + ONLY 1 extra patch that will be taken from the end
    # of the image.

    counter = 0
    original_patch_size_h = patch_size_h
    original_patch_size_w = patch_size_w
    original_step_h = step_h
    original_step_w = step_w

    while (patch_num_h - 1) * step_h + patch_size_h < img_h or (patch_num_h - 2) * step_h + patch_size_h > img_h:
        counter += 1
        if (patch_num_h - 1) * (step_h + 1) + patch_size_h < img_h:
            step_h += 1
        else:
            patch_size_h += 1

        if counter == 100:
            # We probably entered an infinite loop, restart with smaller step size
            step_h = original_step_h - 1
            patch_size_h = original_patch_size_h
            counter = 0

    counter = 0
    while (patch_num_w - 1) * step_w + patch_size_w < img_w or (patch_num_w - 2) * step_w + patch_size_w > img_w:
        counter += 1
        if (patch_num_w - 1) * (step_w + 1) + patch_size_w < img_w:
            step_w += 1
        else:
            patch_size_w += 1

        if counter == 100:
            # We probably entered an infinite loop, restart with smaller step size
            step_w = original_step_w - 1
            patch_size_w = original_patch_size_w
            counter = 0

    return (patch_size_h, patch_size_w), (step_h, step_w)


def compute_patch_info_from_patch_dim(
    img_h: int,
    img_w: int,
    patch_height: int,
    patch_width: int,
    overlap: float = 0.0,
) -> tuple[tuple[int, int], tuple[int, int]]:
    """Compute patch info given the patch dimension
    Args:
        img_h: height of the image
        img_w: width of the image
        patch_height: patch height
        patch_width: patch width
        overlap: overlap percentage [0, 1].

    Returns:
        Tuple of number of patches, step

    """
    assert 1 >= overlap >= 0, f"Invalid overlap. Must be between [0, 1], received {overlap}"
    step_h = patch_height - int(overlap * patch_height)
    step_w = patch_width - int(overlap * patch_width)

    patch_num_h = np.ceil(((img_h - patch_height) / step_h) + 1).astype(int)
    patch_num_w = np.ceil(((img_w - patch_width) / step_w) + 1).astype(int)

    # Handle the case where the last patch does not cover the full image, I need to do this rather than np.ceil
    # because I don't want to add a new patch if the last one exceeds already the image!
    if ((patch_num_h - 1) * step_h) + patch_height < img_h:
        patch_num_h += 1
    if ((patch_num_w - 1) * step_w) + patch_width < img_w:
        patch_num_w += 1

    return (patch_num_h, patch_num_w), (step_h, step_w)


def from_rgb_to_idx(img: np.ndarray, class_to_color: dict, class_to_idx: dict) -> np.ndarray:
    """Args:
        img: Rgb mask in which each different color is associated with a class
        class_to_color: Dict "key": [R, G, B]
        class_to_idx: Dict "key": class_idx.

    Returns:
        Grayscale mask in which each class is associated with a specific index
    """
    img = img.astype(int)
    # Use negative values to avoid strange behaviour in the remote eventuality
    # of someone using a color like [1, 255, 255]
    for classe, color in class_to_color.items():
        img[np.all(img == color, axis=-1).astype(bool), 0] = -class_to_idx[classe]

    img = np.abs(img[:, :, 0])

    return img.astype(np.uint8)


def __save_patch_dataset(
    image_patches: np.ndarray,
    labelled_patches: np.ndarray | None = None,
    mask_patches: np.ndarray | None = None,
    labelled_mask: np.ndarray | None = None,
    output_folder: str = "extraction_data",
    image_name: str = "example",
    area_threshold: float = 0.45,
    area_defect_threshold: float = 0.2,
    mask_extension: str = "_mask",
    save_mask: bool = False,
    mask_output_folder: str | None = None,
    class_to_idx: dict | None = None,
) -> None:
    """Given a view_as_window computed patches, masks and labelled mask, save all the images in subdirectory
    divided by name and position in the grid, ambiguous patches i.e. the one that contains defects but with not enough
    to go above defined thresholds are marked as #DISCARD# and should be discarded in training.
    Patches of images without ground truth are saved inside the None folder.

    Args:
        image_patches: [n, m, patch_w, patch_h, channel] numpy array of the image patches
        mask_patches: [n, m, patch_w, patch_h] numpy array of mask patches
        labelled_patches: [n, m, patch_w, patch_h] numpy array of labelled mask patch
        labelled_mask: numpy array in which each defect in the image is labelled using connected components
        class_to_idx: Dictionary with the mapping {"class" -> class in mask}, it must cover all indices
            contained in the masks
        save_mask: flag to save or ignore mask
        output_folder: folder where to save data
        mask_extension: postfix of the saved mask based on the image name
        mask_output_folder: Optional folder in which to save the masks
        image_name: name to use in order to save the data
        area_threshold: minimum percentage of defected patch area present in the mask to classify the patch as defect
        area_defect_threshold: minimum percentage of single defect present in the patch to classify the patch as defect

    Returns:
        None
    """
    if class_to_idx is not None:
        log.debug("Classes from dict: %s", class_to_idx)
        index_to_class = {v: k for k, v in class_to_idx.items()}
        log.debug("Inverse class: %s", index_to_class)
        reference_classes = index_to_class

        if mask_patches is not None:
            classes_in_mask = set(np.unique(mask_patches))
            missing_classes = set(classes_in_mask).difference(class_to_idx.values())

            assert len(missing_classes) == 0, f"Found index in mask that has no corresponding class {missing_classes}"
    elif mask_patches is not None:
        reference_classes = {k: str(v) for k, v in enumerate(list(np.unique(mask_patches)))}
    else:
        raise ValueError("If no `class_to_idx` is provided, `mask_patches` must be provided")

    log.debug("Classes from mask: %s", reference_classes)
    class_to_idx = {v: k for k, v in reference_classes.items()}
    log.debug("Final reference classes: %s", reference_classes)

    # create subdirectory for the saving data
    for cl in reference_classes.values():
        os.makedirs(os.path.join(output_folder, str(cl)), exist_ok=True)

        if mask_output_folder is not None:
            os.makedirs(os.path.join(output_folder, mask_output_folder, str(cl)), exist_ok=True)

    if mask_output_folder is None:
        mask_output_folder = output_folder
    else:
        mask_output_folder = os.path.join(output_folder, mask_output_folder)

    log.debug("Mask out: %s", mask_output_folder)

    if mask_patches is None:
        os.makedirs(os.path.join(output_folder, str(None)), exist_ok=True)
    # for [i, j] in patches location
    for row_index in range(image_patches.shape[0]):
        for col_index in range(image_patches.shape[1]):
            # default class it's the one in index 0
            output_class = reference_classes.get(0)
            image = image_patches[row_index, col_index]

            discard_in_training = True
            if mask_patches is not None and labelled_patches is not None:
                discard_in_training = False
                max_defected_area = 0
                mask = mask_patches[row_index, col_index]
                patch_area_th = mask.shape[0] * mask.shape[1] * area_threshold

                if mask.sum() > 0:
                    discard_in_training = True
                    for k, v in class_to_idx.items():
                        if v == 0:
                            continue

                        mask_patch = mask == int(v)
                        defected_area = mask_patch.sum()

                        if defected_area > 0:
                            # If enough defected area is inside the patch
                            if defected_area > patch_area_th:
                                if defected_area > max_defected_area:
                                    output_class = k
                                    max_defected_area = defected_area
                                    discard_in_training = False
                            else:
                                all_defects_in_patch = mask_patch * labelled_patches[row_index, col_index]

                                # For each different defect inside the area check
                                # if enough part of it is contained in the patch
                                for defect_id in np.unique(all_defects_in_patch):
                                    if defect_id == 0:
                                        continue

                                    defect_area_in_patch = (all_defects_in_patch == defect_id).sum()
                                    defect_area_th = (labelled_mask == defect_id).sum() * area_defect_threshold

                                    if defect_area_in_patch > defect_area_th:
                                        output_class = k
                                        if defect_area_in_patch > max_defected_area:
                                            max_defected_area = defect_area_in_patch
                                            discard_in_training = False
                        else:
                            discard_in_training = False

                if save_mask:
                    mask_name = f"{image_name}_{row_index * image_patches.shape[1] + col_index}{mask_extension}.png"

                    if discard_in_training:
                        mask_name = "#DISCARD#" + mask_name
                    cv2.imwrite(
                        os.path.join(
                            mask_output_folder,
                            output_class,  # type: ignore[arg-type]
                            mask_name,
                        ),
                        mask.astype(np.uint8),
                    )
            else:
                output_class = "None"

            patch_name = f"{image_name}_{row_index * image_patches.shape[1] + col_index}.png"
            if discard_in_training:
                patch_name = "#DISCARD#" + patch_name

            cv2.imwrite(
                os.path.join(
                    output_folder,
                    output_class,  # type: ignore[arg-type]
                    patch_name,
                ),
                image,
            )


def generate_patch_dataset(
    data_dictionary: list[dict],
    class_to_idx: dict,
    val_size: float = 0.3,
    test_size: float = 0.0,
    seed: int = 42,
    patch_number: tuple[int, int] | None = None,
    patch_size: tuple[int, int] | None = None,
    overlap: float = 0.0,
    output_folder: str = "extraction_data",
    save_original_images_and_masks: bool = True,
    area_threshold: float = 0.45,
    area_defect_threshold: float = 0.2,
    mask_extension: str = "_mask",
    mask_output_folder: str | None = None,
    save_mask: bool = False,
    clear_output_folder: bool = False,
    mask_preprocessing: Callable | None = None,
    train_filename: str = "dataset.txt",
    repeat_good_images: int = 1,
    balance_defects: bool = True,
    annotated_good: list[str] | None = None,
    num_workers: int = 1,
) -> dict | None:
    """Giving a data_dictionary as:
    >>> {
    >>>     'base_name': '163931_1_5.jpg',
    >>>     'path': 'extraction_data/1/163931_1_5.jpg',
    >>>     'mask': 'extraction_data/1/163931_1_5_mask.jpg'
    >>>}
    This function will generate patches datasets based on the defined split number, one for training, one for validation
    and one for testing respectively under output_folder/train, output_folder/val and output_folder/test, the training
    dataset will contain h5 files and a txt file resulting from a call to the
    generate_classification_patch_train_dataset, while the test dataset will contain patches saved on disk divided
    in subfolders per class, patch extraction is done in a sliding window fashion.
    Original images and masks (preprocessed if mask_preprocessing is present) will also be saved under
    output_folder/original/images and output_folder/original/masks.
    If patch number is specified the patch size will be calculated accordingly, if the image is not divisible by the
    patch number two possible behaviours can occur:
        - if the patch reconstruction is smaller than the original image a new patch will be generated containing the
        pixels from the edge of the image (E.g the new patch will contain the last patch_size pixels of the original
        image)
        - if the patch reconstruction is bigger than the original image the last patch will contain the pixels from the
        edge of the image same as above, but without adding a new patch to the count.

    Args:
        data_dictionary: Dictionary as above
        val_size: percentage of the dictionary entries to be used for validation
        test_size: percentage of the dictionary entries to be used for testing
        seed: seed for rng based operations
        clear_output_folder: flag used to delete all the data in subfolder
        class_to_idx: Dictionary {"defect": value in mask.. }
        output_folder: root_folder where to extract the data
        save_original_images_and_masks: If True, images and masks will be copied inside output_folder/original/
        area_threshold: Minimum percentage of defected patch area present in the mask to classify the patch as defect
        area_defect_threshold: Minimum percentage of single defect present in the patch to classify the patch as defect
        mask_extension: Extension used to assign image to mask
        mask_output_folder: Optional folder in which to save the masks
        save_mask: Flag to save the mask
        patch_number: Optional number of patches for each side, required if patch_size is None
        patch_size: Optional dimension of the patch, required if patch_number is None
        overlap: Overlap of the patches [0, 1]
        mask_preprocessing: Optional function applied to masks, this can be useful for example to convert an image in
            range [0-255] to the required [0-1]
        train_filename: Name of the file containing mapping between h5 files and labels for training
        repeat_good_images: Number of repetition for images with emtpy or None mask
        balance_defects: If true add one good entry for each defect extracted
        annotated_good: List of labels that are annotated but considered as good
        num_workers: Number of workers used for the h5 creation

    Returns:
        None if data_dictionary is empty, otherwise return a dictionary containing informations about the dataset

    """
    if len(data_dictionary) == 0:
        warnings.warn("Input data dictionary is empty!", UserWarning, stacklevel=2)
        return None

    if val_size < 0 or test_size < 0 or (val_size + test_size) > 1:
        raise ValueError("Validation and Test size must be greater or equal than zero and sum up to maximum 1")
    if clear_output_folder and os.path.exists(output_folder):
        shutil.rmtree(output_folder)
    os.makedirs(output_folder, exist_ok=True)
    os.makedirs(os.path.join(output_folder, "original"), exist_ok=True)
    if save_original_images_and_masks:
        log.info("Moving original images and masks to dataset folder...")
        os.makedirs(os.path.join(output_folder, "original", "images"), exist_ok=True)
        os.makedirs(os.path.join(output_folder, "original", "masks"), exist_ok=True)

        for i, item in enumerate(data_dictionary):
            img_new_path = os.path.join("original", "images", item["base_name"])
            shutil.copy(item["path"], os.path.join(output_folder, img_new_path))
            data_dictionary[i]["path"] = img_new_path

            if item["mask"] is not None:
                mask = cv2.imread(item["mask"])
                if mask_preprocessing is not None:
                    mask = mask_preprocessing(mask).astype(np.uint8)
                mask_new_path = os.path.join("original", "masks", os.path.splitext(item["base_name"])[0] + ".png")
                cv2.imwrite(os.path.join(output_folder, mask_new_path), mask)
                data_dictionary[i]["mask"] = mask_new_path

    shuffled_indices = np.random.default_rng(seed).permutation(len(data_dictionary))
    data_dictionary = [data_dictionary[i] for i in shuffled_indices]
    log.info("Performing multilabel stratification...")
    train_data_dictionary, val_data_dictionary, test_data_dictionary = multilabel_stratification(
        output_folder=output_folder,
        data_dictionary=data_dictionary,
        num_classes=len(class_to_idx.values()),
        val_size=val_size,
        test_size=test_size,
    )

    log.info("Train set size: %d", len(train_data_dictionary))
    log.info("Validation set size: %d", len(val_data_dictionary))
    log.info("Test set size: %d", len(test_data_dictionary))

    idx_to_class = {v: k for (k, v) in class_to_idx.items()}

    os.makedirs(output_folder, exist_ok=True)

    dataset_info = {
        "patch_size": patch_size,
        "patch_number": patch_number,
        "overlap": overlap,
        "annotated_good": annotated_good,
        "train_files": [{"image_path": x["path"], "mask_path": x["mask"]} for x in train_data_dictionary],
        "val_files": [{"image_path": x["path"], "mask_path": x["mask"]} for x in val_data_dictionary],
        "test_files": [{"image_path": x["path"], "mask_path": x["mask"]} for x in test_data_dictionary],
    }

    with open(os.path.join(output_folder, "info.json"), "w") as f:
        json.dump(dataset_info, f)

    if len(train_data_dictionary) > 0:
        log.info("Generating train set")
        generate_patch_sampling_dataset(
            data_dictionary=train_data_dictionary,
            patch_number=patch_number,
            patch_size=patch_size,
            overlap=overlap,
            idx_to_class=idx_to_class,
            balance_defects=balance_defects,
            repeat_good_images=repeat_good_images,
            output_folder=output_folder,
            subfolder_name="train",
            train_filename=train_filename,
            annotated_good=annotated_good if annotated_good is None else [class_to_idx[x] for x in annotated_good],
            num_workers=num_workers,
        )

    for phase, split_dict in zip(["val", "test"], [val_data_dictionary, test_data_dictionary], strict=False):
        if len(split_dict) > 0:
            log.info("Generating %s set", phase)
            generate_patch_sliding_window_dataset(
                data_dictionary=split_dict,
                patch_number=patch_number,
                patch_size=patch_size,
                overlap=overlap,
                output_folder=output_folder,
                subfolder_name=phase,
                area_threshold=area_threshold,
                area_defect_threshold=area_defect_threshold,
                mask_extension=mask_extension,
                mask_output_folder=mask_output_folder,
                save_mask=save_mask,
                class_to_idx=class_to_idx,
            )

    log.info("All done! Datasets saved to %s", output_folder)

    return dataset_info


def multilabel_stratification(
    output_folder: str,
    data_dictionary: list[dict],
    num_classes: int,
    val_size: float,
    test_size: float,
) -> tuple[list[dict], list[dict], list[dict]]:
    """Split data dictionary using multilabel based stratification, place every sample with None
        mask inside the test set,for all the others read the labels contained in the masks
        to create one-hot encoded labels.

    Args:
        output_folder: root folder of the dataset
        data_dictionary: Data dictionary as described in generate patch dataset
        num_classes: Number of classes contained in the dataset, required for one hot encoding
        val_size: Percentage of data to be used for validation
        test_size: Percentage of data to be used for test
    Returns:
        Three data dictionaries, one for training, one for validation and one for test

    """
    if val_size + test_size == 0:
        return data_dictionary, [], []
    if val_size == 1:
        return [], data_dictionary, []
    if test_size == 1:
        return [], [], data_dictionary

    test_data_dictionary = list(filter(lambda q: q["mask"] is None, data_dictionary))
    log.info("Number of images with no mask inserted in test_data_dictionary: %d", len(test_data_dictionary))
    empty_test_size = len(test_data_dictionary) / len(data_dictionary)
    data_dictionary = list(filter(lambda q: q["mask"] is not None, data_dictionary))

    if len(data_dictionary) == 0:
        # All the item in the data dictionary have None mask, put everything in test
        warnings.warn(
            "All the images have None mask and the test size is not equal to 1! Put everything in test",
            UserWarning,
            stacklevel=2,
        )
        return [], [], test_data_dictionary

    x = []
    y = None
    for item in data_dictionary:
        one_hot = np.zeros([1, num_classes], dtype=np.int16)
        if item["mask"] is None:
            continue
        # this works even if item["mask"] is already an absolute path
        mask = cv2.imread(os.path.join(output_folder, item["mask"]), 0)

        labels = np.unique(mask)

        one_hot[:, labels] = 1
        x.append(item["base_name"])
        if y is None:
            y = one_hot
        else:
            y = np.concatenate([y, one_hot])

    x_test: list[Any] | np.ndarray

    if empty_test_size > test_size:
        warnings.warn(
            (
                "The percentage of images with None label is greater than the test_size, the newest test_size is"
                f" {empty_test_size}!"
            ),
            UserWarning,
            stacklevel=2,
        )
        x_train, _, x_val, _ = iterative_train_test_split(np.expand_dims(np.array(x), 1), y, val_size)
        x_test = [q["base_name"] for q in test_data_dictionary]
    else:
        test_size -= empty_test_size
        x_train, _, x_remaining, y_remaining = iterative_train_test_split(
            np.expand_dims(np.array(x), 1), y, val_size + test_size
        )

        if x_remaining.shape[0] == 1:
            if test_size == 0:
                x_val = x_remaining
                x_test = np.array([])
            elif val_size == 0:
                x_test = x_remaining
                x_val = np.array([])
            else:
                log.warning("Not enough data to create the test split, only a validation set of size 1 will be created")
                x_val = x_remaining
                x_test = np.array([])
        else:
            x_val, _, x_test, _ = iterative_train_test_split(
                x_remaining, y_remaining, test_size / (val_size + test_size)
            )
        # Here x_test should be always a numpy array, but mypy does not recognize it
        x_test = [q[0] for q in x_test.tolist()]  # type: ignore[union-attr]
        x_test.extend([q["base_name"] for q in test_data_dictionary])

    train_data_dictionary = list(filter(lambda q: q["base_name"] in x_train, data_dictionary))
    val_data_dictionary = list(filter(lambda q: q["base_name"] in x_val, data_dictionary))
    test_data_dictionary = list(filter(lambda q: q["base_name"] in x_test, data_dictionary + test_data_dictionary))

    return train_data_dictionary, val_data_dictionary, test_data_dictionary


def generate_patch_sliding_window_dataset(
    data_dictionary: list[dict],
    subfolder_name: str,
    patch_number: tuple[int, int] | None = None,
    patch_size: tuple[int, int] | None = None,
    overlap: float = 0.0,
    output_folder: str = "extraction_data",
    area_threshold: float = 0.45,
    area_defect_threshold: float = 0.2,
    mask_extension: str = "_mask",
    mask_output_folder: str | None = None,
    save_mask: bool = False,
    class_to_idx: dict | None = None,
) -> None:
    """Giving a data_dictionary as:
    >>> {
    >>>     'base_name': '163931_1_5.jpg',
    >>>     'path': 'extraction_data/1/163931_1_5.jpg',
    >>>     'mask': 'extraction_data/1/163931_1_5_mask.jpg'
    >>>}
    This function will extract the patches and save the file and the mask in subdirectory
    Args:
        data_dictionary: Dictionary as above
        subfolder_name: Name of the subfolder where to save the extracted patches (output_folder/subfolder_name)
        class_to_idx: Dictionary {"defect": value in mask.. }
        output_folder: root_folder where to extract the data
        area_threshold: minimum percentage of defected patch area present in the mask to classify the patch as defect
        area_defect_threshold: minimum percentage of single defect present in the patch to classify the patch as defect
        mask_extension: extension used to assign image to mask
        mask_output_folder: Optional folder in which to save the masks
        save_mask: flag to save the mask
        patch_number: Optional number of patches for each side, required if patch_size is None
        patch_size: Optional dimension of the patch, required if patch_number is None
        overlap: overlap of the patches [0, 1].

    Returns:
        None.

    """
    if save_mask and len(mask_extension) == 0 and mask_output_folder is None:
        raise InvalidParameterCombinationException(
            "If mask output folder is not set you must specify a mask extension in order to save masks!"
        )

    if patch_number is None and patch_size is None:
        raise InvalidParameterCombinationException("One between patch number or patch size must be specified!")

    for data in tqdm(data_dictionary):
        base_id = data.get("base_name")
        base_path = data.get("path")
        base_mask = data.get("mask")

        assert base_id is not None, "Cannot find base id in data_dictionary"
        assert base_path is not None, "Cannot find image in data_dictionary"

        image = cv2.imread(os.path.join(output_folder, base_path))
        h = image.shape[0]
        w = image.shape[1]

        log.debug("Processing %s with shape %s", base_id, image.shape)
        mask = mask_patches = None
        labelled_mask = labelled_patches = None

        if base_mask is not None:
            mask = cv2.imread(os.path.join(output_folder, base_mask), 0)
            labelled_mask = label(mask)

        if patch_size is not None:
            [patch_height, patch_width] = patch_size
            [patch_num_h, patch_num_w], step = compute_patch_info_from_patch_dim(
                h, w, patch_height, patch_width, overlap
            )
        elif patch_number is not None:
            [patch_height, patch_width], step = compute_patch_info(h, w, patch_number[0], patch_number[1], overlap)
            [patch_num_h, patch_num_w] = patch_number
        else:
            # mypy does not recognize that this is unreachable
            raise InvalidParameterCombinationException("One between patch number or patch size must be specified!")

        log.debug(
            "Extracting %s patches with size %s, step %s", [patch_num_h, patch_num_w], [patch_height, patch_width], step
        )
        image_patches = extract_patches(image, (patch_num_h, patch_num_w), (patch_height, patch_width), step, overlap)

        if mask is not None:
            if labelled_mask is None:
                raise ValueError("Labelled mask cannot be None!")
            mask_patches = extract_patches(mask, (patch_num_h, patch_num_w), (patch_height, patch_width), step, overlap)
            labelled_patches = extract_patches(
                labelled_mask, (patch_num_h, patch_num_w), (patch_height, patch_width), step, overlap
            )
            assert image_patches.shape[:-1] == mask_patches.shape, "Image patches and mask patches mismatch!"

        log.debug("Image patches shape: %s", image_patches.shape)
        __save_patch_dataset(
            image_patches=image_patches,
            mask_patches=mask_patches,
            labelled_patches=labelled_patches,
            labelled_mask=labelled_mask,
            image_name=os.path.splitext(base_id)[0],
            output_folder=os.path.join(output_folder, subfolder_name),
            area_threshold=area_threshold,
            area_defect_threshold=area_defect_threshold,
            mask_extension=mask_extension,
            save_mask=save_mask,
            mask_output_folder=mask_output_folder,
            class_to_idx=class_to_idx,
        )


def extract_patches(
    image: np.ndarray,
    patch_number: tuple[int, ...],
    patch_size: tuple[int, ...],
    step: tuple[int, ...],
    overlap: float,
) -> np.ndarray:
    """From an image extract N x M Patch[h, w] if the image is not perfectly divided by the number of patches of given
    dimension the last patch will contain a replica of the original image taken in range [-img_h:, :] or [:, -img_w:].

    Args:
        image: Numpy array of the image
        patch_number: number of patches to be extracted
        patch_size: dimension of the patch
        step: step of the patch extraction
        overlap: horizontal and vertical patch overlapping in range [0, 1]

    Returns:
        Patches [N, M, 1, image_w, image_h, image_c]

    """
    assert 1.0 >= overlap >= 0.0, f"Overlap must be between 0 and 1. Received {overlap}"
    (patch_num_h, patch_num_w) = patch_number
    (patch_height, patch_width) = patch_size

    pad_h = (patch_num_h - 1) * step[0] + patch_size[0] - image.shape[0]
    pad_w = (patch_num_w - 1) * step[1] + patch_size[1] - image.shape[1]
    # if the image has 3 channel change dimension
    if len(image.shape) == 3:
        patch_size = (patch_size[0], patch_size[1], image.shape[2])
        step = (step[0], step[1], image.shape[2])

    # If this is not true there's some strange case I didn't take into account
    if pad_h < 0 or pad_w < 0:
        raise ValueError("Something went wrong with the patch extraction, expected positive padding values")

    if pad_h > 0 or pad_w > 0:
        # We work with copies as view_as_windows returns a view of the original image
        crop_img = deepcopy(image)

        if pad_h:
            crop_img = crop_img[0 : (patch_num_h - 2) * step[0] + patch_height, :]

        if pad_w:
            crop_img = crop_img[:, 0 : (patch_num_w - 2) * step[1] + patch_width]

        # Extract safe patches inside the image
        patches = view_as_windows(crop_img, patch_size, step=step)
    else:
        patches = view_as_windows(image, patch_size, step=step)

    extra_patches_h = None
    extra_patches_w = None

    if pad_h > 0:
        # Append extra patches taken from the edge of the image
        extra_patches_h = view_as_windows(image[-patch_height:, :], patch_size, step=step)

    if pad_w > 0:
        extra_patches_w = view_as_windows(image[:, -patch_width:], patch_size, step=step)

        if extra_patches_h is not None:
            # Add an extra column and set is content to the bottom right patch area of the original image if both
            # dimension requires extra patches
            if extra_patches_h.ndim == 6:
                # RGB
                extra_patches_h = np.concatenate(
                    [
                        extra_patches_h,
                        (np.zeros([1, 1, 1, patch_size[0], patch_size[1], extra_patches_h.shape[-1]], dtype=np.uint8)),
                    ],
                    axis=1,
                )
            else:
                extra_patches_h = np.concatenate(
                    [extra_patches_h, (np.zeros([1, 1, patch_size[0], patch_size[1]], dtype=np.uint8))], axis=1
                )

            if extra_patches_h is None:
                # Required by mypy as it cannot infer that extra_patch_h cannot be None
                raise ValueError("Extra patch h cannot be None!")

            extra_patches_h[:, -1, :] = image[-patch_height:, -patch_width:]

    if patches.ndim == 6:
        # With RGB images there's an extra dimension, axis 2 is important don't use plain squeeze or it breaks if
        # the number of patches is set to 1!
        patches = patches.squeeze(axis=2)

    if extra_patches_w is not None:
        if extra_patches_w.ndim == 6:
            # RGB
            patches = np.concatenate([patches, extra_patches_w.squeeze(2)], axis=1)
        else:
            patches = np.concatenate([patches, extra_patches_w], axis=1)

    if extra_patches_h is not None:
        if extra_patches_h.ndim == 6:
            # RGB
            patches = np.concatenate([patches, extra_patches_h.squeeze(2)], axis=0)
        else:
            patches = np.concatenate([patches, extra_patches_h], axis=0)

    # If this is not true there's some strange case I didn't take into account
    assert patches.shape[0] == patch_num_h and patches.shape[1] == patch_num_w, (
        f"Patch shape {patches.shape} does not match the expected shape {patch_number}"
    )

    return patches


def generate_patch_sampling_dataset(
    data_dictionary: list[dict[Any, Any]],
    output_folder: str,
    idx_to_class: dict,
    overlap: float,
    repeat_good_images: int = 1,
    balance_defects: bool = True,
    patch_number: tuple[int, int] | None = None,
    patch_size: tuple[int, int] | None = None,
    subfolder_name: str = "train",
    train_filename: str = "dataset.txt",
    annotated_good: list[int] | None = None,
    num_workers: int = 1,
) -> None:
    """Generate a dataset of patches.

    Args:
        data_dictionary: Dictionary containing image and mask mapping
        output_folder: root folder
        idx_to_class: Dict mapping an index to the corresponding class name
        repeat_good_images: Number of repetition for images with emtpy or None mask
        balance_defects: If true add one good entry for each defect extracted
        patch_number: Optional number of patches for each side, required if patch_size is None
        patch_size: Optional dimension of the patch, required if patch_number is None
        overlap: Percentage of overlap between patches
        subfolder_name: name of the subfolder where to store h5 files for defected images and dataset txt
        train_filename: Name of the file in which to store the mappings between h5 files and labels
        annotated_good: List of class indices that are considered good other than the background
        num_workers: Number of processes used to create h5 files.

    Returns:
        Create a txt file containing tuples path,label where path is a pointer to the generated h5 file and label is the
            corresponding label

            Each generated h5 file contains five fields:
                img_path: Pointer to the location of the original image
                mask_path: Optional pointer to the mask file, is missing if the mask is completely empty or is
                not present
                patch_size: dimension of the patches on the interested image
                triangles: List of triangles that covers the defect
                triangles_weights: Which weight should be given to each triangle for sampling

    """
    if patch_number is None and patch_size is None:
        raise InvalidParameterCombinationException("One between patch number or patch size must be specified!")

    sampling_dataset_folder = os.path.join(output_folder, subfolder_name)

    os.makedirs(sampling_dataset_folder, exist_ok=True)
    labelled_masks_path = os.path.join(output_folder, "original", "labelled_masks")
    os.makedirs(labelled_masks_path, exist_ok=True)

    with open(os.path.join(sampling_dataset_folder, train_filename), "w") as output_file:
        if num_workers < 1:
            raise InvalidNumWorkersNumberException("Workers must be >= 1")

        if num_workers > 1:
            log.info("Executing generate_patch_sampling_dataset w/ more than 1 worker!")

            split_data_dictionary = np.array_split(np.asarray(data_dictionary), num_workers)

            with Pool(num_workers) as pool:
                res_list = pool.map(
                    partial(
                        create_h5,
                        patch_size=patch_size,
                        patch_number=patch_number,
                        idx_to_class=idx_to_class,
                        overlap=overlap,
                        repeat_good_images=repeat_good_images,
                        balance_defects=balance_defects,
                        annotated_good=annotated_good,
                        output_folder=output_folder,
                        labelled_masks_path=labelled_masks_path,
                        sampling_dataset_folder=sampling_dataset_folder,
                    ),
                    split_data_dictionary,
                )

            res = list(itertools.chain(*res_list))
        else:
            res = create_h5(
                data_dictionary=data_dictionary,
                patch_size=patch_size,
                patch_number=patch_number,
                idx_to_class=idx_to_class,
                overlap=overlap,
                repeat_good_images=repeat_good_images,
                balance_defects=balance_defects,
                annotated_good=annotated_good,
                output_folder=output_folder,
                labelled_masks_path=labelled_masks_path,
                sampling_dataset_folder=sampling_dataset_folder,
            )

        for line in res:
            output_file.write(line)


def create_h5(
    data_dictionary: list[dict[Any, Any]],
    idx_to_class: dict,
    overlap: float,
    repeat_good_images: int,
    balance_defects: bool,
    output_folder: str,
    labelled_masks_path: str,
    sampling_dataset_folder: str,
    annotated_good: list[int] | None = None,
    patch_size: tuple[int, int] | None = None,
    patch_number: tuple[int, int] | None = None,
) -> list[str]:
    """Create h5 files for each image in the dataset.

    Args:
        data_dictionary: Dictionary containing image and mask mapping
        idx_to_class: Dict mapping an index to the corresponding class name
        overlap: Percentage of overlap between patches
        repeat_good_images: Number of repetition for images with emtpy or None mask
        balance_defects: If true add one good entry for each defect extracted
        output_folder: root folder
        overlap: Percentage of overlap between patches
        annotated_good: List of class indices that are considered good other than the background
        labelled_masks_path: paths of labelled masks
        sampling_dataset_folder: folder of the dataset
        patch_size: Dimension of the patch, required if patch_number is None
        patch_number: Number of patches for each side, required if patch_size is None.

    Returns:
        output_list: List of h5 files' names

    """
    if patch_number is None and patch_size is None:
        raise InvalidParameterCombinationException("One between patch number or patch size must be specified!")

    output_list = []
    for item in tqdm(data_dictionary):
        log.debug("Processing %s", item["base_name"])
        # this works even if item["path"] is already an absolute path
        img = cv2.imread(os.path.join(output_folder, item["path"]))

        h = img.shape[0]
        w = img.shape[1]

        mask: np.ndarray
        if item["mask"] is None:
            mask = np.zeros([h, w], dtype=np.uint8)
        else:
            # this works even if item["mask"] is already an absolute path
            mask = cv2.imread(os.path.join(output_folder, item["mask"]), 0)

        if patch_size is not None:
            patch_height = patch_size[1]
            patch_width = patch_size[0]
        else:
            # Mypy complains because patch_number is Optional, but we already checked that it is not None.
            [patch_height, patch_width], _ = compute_patch_info(
                h,
                w,
                patch_number[0],  # type: ignore[index]
                patch_number[1],  # type: ignore[index]
                overlap,
            )

        h5_file_name_good = os.path.join(sampling_dataset_folder, f"{os.path.splitext(item['base_name'])[0]}_good.h5")

        disable_good = False

        with h5py.File(h5_file_name_good, "w") as f:
            f.create_dataset("img_path", data=item["path"])
            f.create_dataset("patch_size", data=np.array([patch_height, patch_width]))

            target = idx_to_class[0]

            if mask.sum() == 0:
                f.create_dataset("triangles", data=np.array([], dtype=np.uint8), dtype=np.uint8)
                f.create_dataset("triangles_weights", data=np.array([], dtype=np.uint8), dtype=np.uint8)

                for _ in range(repeat_good_images):
                    output_list.append(f"{os.path.basename(h5_file_name_good)},{target}\n")

                continue

            binary_mask = (mask > 0).astype(np.uint8)

            # Dilate the defects and take the background
            binary_mask = np.logical_not(cv2.dilate(binary_mask, np.ones([patch_height, patch_width]))).astype(np.uint8)

            temp_binary_mask = deepcopy(binary_mask)
            # Remove the edges of the image as they are unsafe for sampling without padding
            temp_binary_mask[0 : patch_height // 2, :] = 0
            temp_binary_mask[:, 0 : patch_width // 2] = 0
            temp_binary_mask[-patch_height // 2 :, :] = 0
            temp_binary_mask[:, -patch_width // 2 :] = 0

            if temp_binary_mask.sum() != 0:
                # If the mask without the edges is not empty use it, otherwise use the original mask as it is not
                # possible to sample a patch that will not exceed the edges, this must be taken care by the patch
                # sampler used during training
                binary_mask = temp_binary_mask

            # In the case of hx1 or 1xw number of patches we must make sure that the sampling row or the sampling
            # column is empty, if it isn't remove it from the possible sampling area
            if patch_height == img.shape[0]:
                must_clear_indices = np.where(binary_mask.sum(axis=0) != img.shape[0])[0]
                binary_mask[:, must_clear_indices] = 0

            if patch_width == img.shape[1]:
                must_clear_indices = np.where(binary_mask.sum(axis=1) != img.shape[1])[0]
                binary_mask[must_clear_indices, :] = 0

            # If there's no space for sampling good patches skip it
            if binary_mask.sum() == 0:
                disable_good = True
            else:
                triangles, weights = triangulate_region(binary_mask)
                if triangles is None:
                    disable_good = True
                else:
                    log.debug(
                        "Saving %s triangles for %s with label %s",
                        triangles.shape[0],
                        os.path.basename(h5_file_name_good),
                        target,
                    )

                    f.create_dataset("mask_path", data=item["mask"])
                    # Points from extracted triangles should be sufficiently far from all the defects allowing to sample
                    # good patches almost all the time
                    f.create_dataset("triangles", data=triangles, dtype=np.int32)
                    f.create_dataset("triangles_weights", data=weights, dtype=np.float64)

                    # Avoid saving the good h5 file here because otherwise I'll have one more good compared to the
                    # number of defects
                    if not balance_defects:
                        output_list.append(f"{os.path.basename(h5_file_name_good)},{target}\n")

        if disable_good:
            os.remove(h5_file_name_good)

        labelled_mask = label(mask)
        cv2.imwrite(os.path.join(labelled_masks_path, f"{os.path.splitext(item['base_name'])[0]}.png"), labelled_mask)

        real_defects_mask = None

        if annotated_good is not None:
            # Remove true defected area from the good labeled mask
            # If we want this to be even more restrictive we could also include the background as we don't know for sure
            # it will not contain any defects
            real_defects_mask = (~np.isin(mask, [0] + annotated_good)).astype(np.uint8)
            real_defects_mask = cv2.dilate(real_defects_mask, np.ones([patch_height, patch_width])).astype(bool)

        for i in np.unique(labelled_mask):
            if i == 0:
                continue

            current_mask = (labelled_mask == i).astype(np.uint8)
            target_idx = (mask * current_mask).max()

            # When we have good annotations we want to avoid sampling patches containing true defects, to do so we
            # reduce the extraction area based on the area covered by the other defects
            if annotated_good is not None and real_defects_mask is not None and target_idx in annotated_good:
                # a - b = a & ~b
                # pylint: disable=invalid-unary-operand-type
                current_mask = np.bitwise_and(current_mask.astype(bool), ~real_defects_mask).astype(np.uint8)
            else:
                # When dealing with small defects the number of points that will be sampled will be limited and patches
                # will mostly be centered around the defect, to overcome this issue enlarge defect bounding box by 50%
                # of the difference between the patch_size and the defect bb size, we don't do this on good labels to
                # avoid invalidating the reduction applied before.
                props = regionprops(current_mask)[0]
                bbox_size = [props.bbox[2] - props.bbox[0], props.bbox[3] - props.bbox[1]]
                diff_bbox = np.array([max(0, patch_height - bbox_size[0]), max(0, patch_width - bbox_size[1])])

                if diff_bbox[0] != 0:
                    current_mask = cv2.dilate(current_mask, np.ones([diff_bbox[0] // 2, 1]))
                if diff_bbox[1] != 0:
                    current_mask = cv2.dilate(current_mask, np.ones([1, diff_bbox[1] // 2]))

            if current_mask.sum() == 0:
                # If it's not possible to sample a labelled good patch basically
                continue

            temp_current_mask = deepcopy(current_mask)
            # Remove the edges of the image as they are unsafe for sampling without padding
            temp_current_mask[0 : patch_height // 2, :] = 0
            temp_current_mask[:, 0 : patch_width // 2] = 0
            temp_current_mask[-patch_height // 2 :, :] = 0
            temp_current_mask[:, -patch_width // 2 :] = 0

            if temp_current_mask.sum() != 0:
                # If the mask without the edges is not empty use it, otherwise use the original mask as it is not
                # possible to sample a patch that will not exceed the edges, this must be taken care by the patch
                # sampler used during training
                current_mask = temp_current_mask

            triangles, weights = triangulate_region(current_mask)

            if triangles is not None:
                h5_file_name = os.path.join(sampling_dataset_folder, f"{os.path.splitext(item['base_name'])[0]}_{i}.h5")

                target = idx_to_class[target_idx]

                log.debug(
                    "Saving %s triangles for %s with label %s",
                    triangles.shape[0],
                    os.path.basename(h5_file_name),
                    target,
                )

                with h5py.File(h5_file_name, "w") as f:
                    f.create_dataset("img_path", data=item["path"])
                    f.create_dataset("mask_path", data=item["mask"])
                    f.create_dataset("patch_size", data=np.array([patch_height, patch_width]))
                    f.create_dataset("triangles", data=triangles, dtype=np.int32)
                    f.create_dataset("triangles_weights", data=weights, dtype=np.float64)
                    f.create_dataset("labelled_index", data=i, dtype=np.int32)

                if annotated_good is not None and target_idx in annotated_good:
                    # I treat annotate good images exactly the same as I would treat background
                    for _ in range(repeat_good_images):
                        output_list.append(f"{os.path.basename(h5_file_name)},{target}\n")
                else:
                    output_list.append(f"{os.path.basename(h5_file_name)},{target}\n")

                if balance_defects:
                    if not disable_good:
                        output_list.append(f"{os.path.basename(h5_file_name_good)},{idx_to_class[0]}\n")
                    else:
                        log.debug(
                            "Unable to add a good defect for %s, since there's no way to sample good patches",
                            h5_file_name,
                        )
    return output_list


def triangle_area(triangle: np.ndarray) -> float:
    """Compute the area of a triangle defined by 3 points.

    Args:
        triangle: Array of shape 3x2 containing the coordinates of a triangle.

    Returns:
        The area of the triangle

    """
    [y1, x1], [y2, x2], [y3, x3] = triangle
    return abs(0.5 * (((x2 - x1) * (y3 - y1)) - ((x3 - x1) * (y2 - y1))))


def triangulate_region(mask: ndimage) -> tuple[np.ndarray | None, np.ndarray | None]:
    """Extract from a binary image containing a single roi (with or without holes) a list of triangles
    (and their normalized area) that completely subdivide an approximated polygon defined around mask contours,
    the output can be used to easily sample uniformly points that are almost guarantee to lie inside the roi.

    Args:
        mask: Binary image defining a region of interest

    Returns:
        Tuple containing:
            triangles: a numpy array containing a list of list of vertices (y, x) of the triangles defined over a
                polygon that contains the entire region
            weights: areas of each triangle rescaled (area_i / sum(areas))

    """
    polygon_points, hier = cv2.findContours(mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_TC89_L1)

    if not np.all(hier[:, :, 3] == -1):  # there are holes
        holes = ndimage.binary_fill_holes(mask).astype(np.uint8)
        holes -= mask
        holes = (holes > 0).astype(np.uint8)
        if holes.sum() > 0:  # there are holes
            for hole in regionprops(label(holes)):
                y_hole_center = int(hole.centroid[0])
                mask[y_hole_center] = 0

        polygon_points, _ = cv2.findContours(mask, cv2.RETR_LIST, cv2.CHAIN_APPROX_TC89_L1)

    final_approx = []

    # Extract a simpler approximation of the contour
    for cnt in polygon_points:
        epsilon = 0.01 * cv2.arcLength(cnt, True)
        approx = cv2.approxPolyDP(cnt, epsilon, True)
        final_approx.append(approx)

    triangles = None

    for approx in final_approx:
        contours_tripy = [x[0] for x in approx]
        current_triangles = earclip(contours_tripy)

        if len(current_triangles) == 0:
            # This can only happen is a defect is like one pixel wide...
            continue

        current_triangles = np.array([list(x) for x in current_triangles])

        triangles = current_triangles if triangles is None else np.concatenate([triangles, current_triangles])

    if triangles is None:
        return None, None

    # Swap x and y to match cv2
    triangles = triangles[..., ::-1]

    weights = np.array([triangle_area(x) for x in triangles])
    weights = weights / weights.sum()

    return triangles, weights


class InvalidParameterCombinationException(Exception):
    """Exception raised when an invalid combination of parameters is passed to a function."""


class InvalidNumWorkersNumberException(Exception):
    """Exception raised when an invalid number of workers is passed to a function."""


def load_train_file(
    train_file_path: str,
    include_filter: list[str] | None = None,
    exclude_filter: list[str] | None = None,
    class_to_skip: list | None = None,
) -> tuple[list[str], list[str]]:
    """Load a train file and return a list of samples and a list of targets. It is expected that train files will be in
        the same location as the train_file_path.

    Args:
        train_file_path: Training file location
        include_filter: Include only samples that contain one of the element of this list
        exclude_filter: Exclude all samples that contain one of the element of this list
        class_to_skip: if not None, exlude all the samples with labels present in this list.

    Returns:
        List of samples and list of targets

    """
    samples = []
    targets = []

    with open(train_file_path) as f:
        lines = f.read().splitlines()
        for line in lines:
            sample, target = line.split(",")
            if class_to_skip is not None and target in class_to_skip:
                continue
            samples.append(sample)
            targets.append(target)

    include_filter = [] if include_filter is None else include_filter
    exclude_filter = [] if exclude_filter is None else exclude_filter

    valid_samples_indices = [
        i
        for (i, x) in enumerate(samples)
        if (len(include_filter) == 0 or any(f in x for f in include_filter))
        and (len(exclude_filter) == 0 or not any(f in x for f in exclude_filter))
    ]

    samples = [samples[i] for i in valid_samples_indices]
    targets = [targets[i] for i in valid_samples_indices]

    train_folder = os.path.dirname(train_file_path)
    samples = [os.path.join(train_folder, x) for x in samples]

    return samples, targets


def compute_safe_patch_range(sampled_point: int, patch_size: int, image_size: int) -> tuple[int, int]:
    """Computes the safe patch size for the given image size.

    Args:
        sampled_point: the sampled point
        patch_size: the size of the patch
        image_size: the size of the image.

    Returns:
        Tuple containing the safe patch range [left, right] such that
        [sampled_point - left : sampled_point + right] will be within the image size.
    """
    left = patch_size // 2
    right = patch_size // 2

    if sampled_point + right > image_size:
        right = image_size - sampled_point
        left = patch_size - right

    if sampled_point - left < 0:
        left = sampled_point
        right = patch_size - left

    return left, right


def trisample(triangle: np.ndarray) -> tuple[int, int]:
    """Sample a point uniformly in a triangle.

    Args:
        triangle: Array of shape 3x2 containing the coordinates of a triangle.

    Returns:
        Sample point uniformly in the triangle

    """
    [y1, x1], [y2, x2], [y3, x3] = triangle

    r1 = random.random()
    r2 = random.random()

    s1 = math.sqrt(r1)

    x = x1 * (1.0 - s1) + x2 * (1.0 - r2) * s1 + x3 * r2 * s1
    y = y1 * (1.0 - s1) + y2 * (1.0 - r2) * s1 + y3 * r2 * s1

    return int(y), int(x)



from __future__ import annotations

import os
import warnings

import cv2
import numpy as np
import pandas as pd
from scipy import ndimage
from skimage.measure import label, regionprops  # pylint: disable=no-name-in-module
from tqdm import tqdm

from quadra.utils import utils
from quadra.utils.patch.dataset import PatchDatasetFileFormat, compute_patch_info, compute_patch_info_from_patch_dim

log = utils.get_logger(__name__)


def get_sorted_patches_by_image(test_results: pd.DataFrame, img_name: str) -> pd.DataFrame:
    """Gets the patches of a given image sorted by patch number.

    Args:
        test_results: Pandas dataframe containing test results like the one produced by SklearnClassificationTrainer
        img_name: name of the image used to filter the results.

    Returns:
        test results filtered by image name and sorted by patch number
    """
    img_patches = test_results[test_results["filename"] == os.path.splitext(img_name)[0]]
    patches_idx = np.array(
        [int(os.path.basename(x).split("_")[-1].replace(".png", "")) for x in img_patches["sample"].tolist()]
    )
    patches_idx = np.argsort(patches_idx).tolist()
    img_patches = img_patches.iloc[patches_idx]

    return img_patches


def compute_patch_metrics(
    test_img_info: list[PatchDatasetFileFormat],
    test_results: pd.DataFrame,
    overlap: float,
    idx_to_class: dict,
    patch_num_h: int | None = None,
    patch_num_w: int | None = None,
    patch_w: int | None = None,
    patch_h: int | None = None,
    return_polygon: bool = False,
    patch_reconstruction_method: str = "priority",
    annotated_good: list[int] | None = None,
) -> tuple[int, int, int, list[dict]]:
    """Compute the metrics of a patch dataset.

    Args:
        test_img_info: List of observation paths and mask paths
        test_results: Pandas dataframe containing the results of an SklearnClassificationTrainer utility
        patch_num_h: Number of vertical patches (required if patch_w and patch_h are None)
        patch_num_w: Number of horizontal patches (required if patch_w and patch_h are None)
        patch_h: Patch height (required if patch_num_h and patch_num_w are None)
        patch_w: Patch width (required if patch_num_h and patch_num_w are None)
        overlap: Percentage of overlap between the patches
        idx_to_class: Dict mapping an index to the corresponding class name
        return_polygon: if set to true convert the reconstructed mask into polygons, otherwise return the mask
        patch_reconstruction_method: How to compute the label of overlapping patches, can either be:
            priority: Assign the top priority label (i.e the one with greater index) to overlapping regions
            major_voting: Assign the most present label among the patches label overlapping a pixel
        annotated_good: List of indices of annotations to be treated as good.

    Returns:
        Tuple containing:
            false_region_bad: Number of false bad regions detected in the dataset
            false_region_good: Number of missed defects
            true_region_bad: Number of correctly identified defects
            reconstructions: If polygon is true this is a List of dict containing
                {
                    "file_path": image_path,
                    "mask_path": mask_path,
                    "file_name": observation_name,
                    "prediction": [{
                        "label": predicted_label,
                        "points": List of dict coordinates "x" and "y" representing the points of a polygon that
                        surrounds an image area covered by patches of label = predicted_label
                    }]
                }
            else its a list of dict containing
                {
                    "file_path": image_path,
                    "mask_path": mask_path,
                    "file_name": observation_name,
                    "prediction": numpy array containing the reconstructed mask
                }
    """
    assert patch_reconstruction_method in [
        "priority",
        "major_voting",
    ], "Patch reconstruction method not recognized, valid values are priority, major_voting"

    if (patch_h is not None and patch_w is not None) and (patch_num_h is not None and patch_num_w is not None):
        raise ValueError("Either number of patches or patch size is required for reconstruction")

    assert (patch_h is not None and patch_w is not None) or (patch_num_h is not None and patch_num_w is not None), (
        "Either number of patches or patch size is required for reconstruction"
    )

    if patch_h is not None and patch_w is not None and patch_num_h is not None and patch_num_w is not None:
        warnings.warn(
            "Both number of patches and patch dimension are specified, using number of patches by default",
            UserWarning,
            stacklevel=2,
        )

    log.info("Computing patch metrics!")

    false_region_bad = 0
    false_region_good = 0
    true_region_bad = 0
    reconstructions = []
    test_results["filename"] = test_results["sample"].apply(
        lambda x: "_".join(os.path.basename(x).replace("#DISCARD#", "").split("_")[0:-1])
    )

    for info in tqdm(test_img_info):
        img_path = info.image_path
        mask_path = info.mask_path

        img_json_entry = {
            "image_path": img_path,
            "mask_path": mask_path,
            "file_name": os.path.basename(img_path),
            "prediction": None,
        }

        test_img = cv2.imread(img_path)

        img_name = os.path.basename(img_path)

        h = test_img.shape[0]
        w = test_img.shape[1]

        gt_img = None

        if mask_path is not None and os.path.exists(mask_path):
            gt_img = cv2.imread(mask_path, 0)
            if test_img.shape[0:2] != gt_img.shape:
                # Ensure that the mask has the same size as the image by padding it with zeros
                log.warning("Found mask with different size than the image, padding it with zeros!")
                gt_img = np.pad(
                    gt_img, ((0, test_img.shape[0] - gt_img.shape[0]), (0, test_img.shape[1] - gt_img.shape[1]))
                )
        if patch_num_h is not None and patch_num_w is not None:
            patch_size, step = compute_patch_info(h, w, patch_num_h, patch_num_w, overlap)
        elif patch_h is not None and patch_w is not None:
            [patch_num_h, patch_num_w], step = compute_patch_info_from_patch_dim(h, w, patch_h, patch_w, overlap)
            patch_size = (patch_h, patch_w)
        else:
            raise ValueError(
                "Either number of patches or patch size is required for reconstruction, this should not happen"
                " at this stage"
            )

        img_patches = get_sorted_patches_by_image(test_results, img_name)
        pred = img_patches["pred_label"].to_numpy().reshape(patch_num_h, patch_num_w)

        # Treat annotated good predictions as background, this is an optimistic assumption that assumes that the
        # remaining background is good, but it is not always true so maybe on non annotated areas we are missing
        # defects and it would be necessary to handle this in a different way.
        if annotated_good is not None:
            pred[np.isin(pred, annotated_good)] = 0
        if patch_num_h is not None and patch_num_w is not None:
            output_mask, predicted_defect = reconstruct_patch(
                input_img_shape=test_img.shape,
                patch_size=patch_size,
                pred=pred,
                patch_num_h=patch_num_h,
                patch_num_w=patch_num_w,
                idx_to_class=idx_to_class,
                step=step,
                return_polygon=return_polygon,
                method=patch_reconstruction_method,
            )
        else:
            raise ValueError("`patch_num_h` and `patch_num_w` cannot be None at this point")

        if return_polygon:
            img_json_entry["prediction"] = predicted_defect
        else:
            img_json_entry["prediction"] = output_mask

        reconstructions.append(img_json_entry)
        if gt_img is not None:
            if annotated_good is not None:
                gt_img[np.isin(gt_img, annotated_good)] = 0

            gt_img_binary = (gt_img > 0).astype(bool)
            regions_pred = label(output_mask).astype(np.uint8)

            for k in range(1, regions_pred.max() + 1):
                region = (regions_pred == k).astype(bool)
                # If there's no overlap with the gt
                if np.sum(np.bitwise_and(region, gt_img_binary)) == 0:
                    false_region_bad += 1

            output_mask = (output_mask > 0).astype(np.uint8)
            gt_img = label(gt_img)

            if gt_img is None:
                raise RuntimeError("Ground truth mask is None after label and it should not be")

            for i in range(1, gt_img.max() + 1):
                region = (gt_img == i).astype(bool)
                if np.sum(np.bitwise_and(region, output_mask)) == 0:
                    false_region_good += 1
                else:
                    true_region_bad += 1

    return false_region_bad, false_region_good, true_region_bad, reconstructions


def reconstruct_patch(
    input_img_shape: tuple[int, ...],
    patch_size: tuple[int, int],
    pred: np.ndarray,
    patch_num_h: int,
    patch_num_w: int,
    idx_to_class: dict,
    step: tuple[int, int],
    return_polygon: bool = True,
    method: str = "priority",
) -> tuple[np.ndarray, list[dict]]:
    """Reconstructs the prediction image from the patches.

    Args:
        input_img_shape: The size of the reconstructed image
        patch_size: Array defining the patch size
        pred: Numpy array containing reconstructed prediction (patch_num_h x patch_num_w)
        patch_num_h: Number of vertical patches
        patch_num_w: Number of horizontal patches
        idx_to_class: Dictionary mapping indices to labels
        step: Array defining the step size to be used for reconstruction
        return_polygon: If true compute predicted polygons. Defaults to True.
        method: Reconstruction method to be used. Currently supported: "priority" and "major_voting"

    Returns:
        (reconstructed_prediction_image, predictions) where predictions is an array of objects
            [{
                "label": Predicted_label,
                "points": List of dict coordinates "x" and "y" representing the points of a polygon that
                    surrounds an image area covered by patches of label = predicted_label
            }]
    """
    if method == "priority":
        return _reconstruct_patch_priority(
            input_img_shape,
            patch_size,
            pred,
            patch_num_h,
            patch_num_w,
            idx_to_class,
            step,
            return_polygon,
        )
    if method == "major_voting":
        return _reconstruct_patch_major_voting(
            input_img_shape,
            patch_size,
            pred,
            patch_num_h,
            patch_num_w,
            idx_to_class,
            step,
            return_polygon,
        )

    raise ValueError(f"Invalid reconstruction method {method}")


def _reconstruct_patch_priority(
    input_img_shape: tuple[int, ...],
    patch_size: tuple[int, int],
    pred: np.ndarray,
    patch_num_h: int,
    patch_num_w: int,
    idx_to_class: dict,
    step: tuple[int, int],
    return_polygon: bool = True,
) -> tuple[np.ndarray, list[dict]]:
    """Reconstruct patch polygons using the priority method."""
    final_mask = np.zeros([input_img_shape[0], input_img_shape[1]], dtype=np.uint8)
    predicted_defect = []

    for i in range(1, pred.max() + 1):
        white_patch = np.full((patch_size[0], patch_size[1]), i, dtype=np.uint8)
        masked_pred = (pred == i).astype(np.uint8)

        if masked_pred.sum() == 0:
            continue

        mask_img = np.zeros([input_img_shape[0], input_img_shape[1]], dtype=np.uint8)

        for h in range(patch_num_h):
            for w in range(patch_num_w):
                if masked_pred[h, w] == 1:
                    patch_location_h = step[0] * h
                    patch_location_w = step[1] * w

                    # Move replicated patches prediction in the correct position of the original image if needed
                    if patch_location_h + patch_size[0] > mask_img.shape[0]:
                        patch_location_h = mask_img.shape[0] - patch_size[0]

                    if patch_location_w + patch_size[1] > mask_img.shape[1]:
                        patch_location_w = mask_img.shape[1] - patch_size[1]

                    mask_img[
                        patch_location_h : patch_location_h + patch_size[0],
                        patch_location_w : patch_location_w + patch_size[1],
                    ] = white_patch

        mask_img = mask_img[0 : input_img_shape[0], 0 : input_img_shape[1]]

        # Priority is given by the index of the class, the larger, the more important
        final_mask = np.maximum(mask_img, final_mask)

    if final_mask.sum() != 0 and return_polygon:
        for lab in np.unique(final_mask):
            if lab == 0:
                continue

            polygon = from_mask_to_polygon((final_mask == lab).astype(np.uint8))

            for pol in polygon:
                class_entry = {
                    "label": idx_to_class.get(lab),
                    "points": pol,
                }

                predicted_defect.append(class_entry)

    return final_mask, predicted_defect


def _reconstruct_patch_major_voting(
    input_img_shape: tuple[int, ...],
    patch_size: tuple[int, int],
    pred: np.ndarray,
    patch_num_h: int,
    patch_num_w: int,
    idx_to_class: dict,
    step: tuple[int, int],
    return_polygon: bool = True,
):
    """Reconstruct patch polygons using the major voting method."""
    predicted_defect = []

    final_mask = np.zeros([input_img_shape[0], input_img_shape[1], np.max(pred) + 1], dtype=np.uint8)
    white_patch = np.ones((patch_size[0], patch_size[1]), dtype=np.uint8)

    for i in range(1, pred.max() + 1):
        masked_pred = (pred == i).astype(np.uint8)

        if masked_pred.sum() == 0:
            continue

        mask_img = np.zeros([input_img_shape[0], input_img_shape[1]], dtype=np.uint8)

        for h in range(patch_num_h):
            for w in range(patch_num_w):
                if masked_pred[h, w] == 1:
                    patch_location_h = step[0] * h
                    patch_location_w = step[1] * w

                    # Move replicated patches prediction in the correct position of the original image if needed
                    if patch_location_h + patch_size[0] > mask_img.shape[0]:
                        patch_location_h = mask_img.shape[0] - patch_size[0]

                    if patch_location_w + patch_size[1] > mask_img.shape[1]:
                        patch_location_w = mask_img.shape[1] - patch_size[1]

                    mask_img[
                        patch_location_h : patch_location_h + patch_size[0],
                        patch_location_w : patch_location_w + patch_size[1],
                    ] += white_patch

        mask_img = mask_img[0 : input_img_shape[0], 0 : input_img_shape[1]]
        final_mask[:, :, i] = mask_img

    # Since argmax returns first element on ties and the priority is defined from 0 to n_classes,
    # I needed a way to get the last element on ties, this code achieves that
    final_mask = ((final_mask.shape[-1] - 1) - np.argmax(final_mask[..., ::-1], axis=-1)) * np.invert(
        np.all(final_mask == 0, axis=-1)
    )

    if final_mask.sum() != 0 and return_polygon:
        for lab in np.unique(final_mask):
            if lab == 0:
                continue

            polygon = from_mask_to_polygon((final_mask == lab).astype(np.uint8))

            for pol in polygon:
                class_entry = {
                    "label": idx_to_class.get(lab),
                    "points": pol,
                }

                predicted_defect.append(class_entry)

    return final_mask, predicted_defect


def from_mask_to_polygon(mask_img: np.ndarray) -> list:
    """Convert a mask of pattern to a list of polygon vertices.

    Args:
        mask_img: masked patch reconstruction image
    Returns:
        a list of lists containing the coordinates of the polygons containing each region of the mask:
        [
            [
                {
                    "x": 1.1,
                    "y": 2.2
                },
                {
                    "x": 2.1,
                    "y": 3.2
                }
            ], ...
        ].
    """
    points_dict = []
    # find vertices of polygon: points -> list of array of dim n_vertex, 1, 2(x,y)
    polygon_points, hier = cv2.findContours(mask_img, cv2.RETR_TREE, cv2.CHAIN_APPROX_TC89_L1)

    if not hier[:, :, 2:].all(-1).all():  # there are holes
        holes = ndimage.binary_fill_holes(mask_img).astype(int)
        holes -= mask_img
        holes = (holes > 0).astype(np.uint8)
        if holes.sum() > 0:  # there are holes
            for hole in regionprops(label(holes)):
                a, _, _, _d = hole.bbox
                mask_img[a] = 0

        polygon_points, hier = cv2.findContours(mask_img, cv2.RETR_LIST, cv2.CHAIN_APPROX_TC89_L1)

    for pol in polygon_points:
        #  pol: n_vertex, 1, 2
        current_poly = []
        for point in pol:
            current_poly.append({"x": int(point[0, 0]), "y": int(point[0, 1])})
        points_dict.append(current_poly)

    return points_dict



from __future__ import annotations

import shutil
from dataclasses import dataclass
from pathlib import Path
from typing import Any

import cv2
import pytest

from quadra.utils.tests.helpers import _random_image


@dataclass
class AnomalyDatasetArguments:
    """Anomaly dataset arguments.

    Args:
        train_samples: number of train samples
        val_samples: number of validation samples (good, bad)
        test_samples: number of test samples (good, bad)
    """

    train_samples: int
    val_samples: tuple[int, int]
    test_samples: tuple[int, int]


def _build_anomaly_dataset(
    tmp_path: Path, dataset_arguments: AnomalyDatasetArguments
) -> tuple[str, AnomalyDatasetArguments]:
    """Generate anomaly dataset in the standard mvtec format.

    Args:
        tmp_path: path to temporary directory
        dataset_arguments: dataset arguments

    Returns:
        path to anomaly dataset
    """
    train_samples = dataset_arguments.train_samples
    val_samples = dataset_arguments.val_samples
    test_samples = dataset_arguments.test_samples

    anomaly_dataset_path = tmp_path / "anomaly_dataset"
    anomaly_dataset_path.mkdir()
    train_good_path = anomaly_dataset_path / "train" / "good"
    val_good_path = anomaly_dataset_path / "val" / "good"
    val_bad_path = anomaly_dataset_path / "val" / "bad"
    test_good_path = anomaly_dataset_path / "test" / "good"
    test_bad_path = anomaly_dataset_path / "test" / "bad"

    train_good_path.mkdir(parents=True)
    val_good_path.mkdir(parents=True)
    val_bad_path.mkdir(parents=True)
    test_good_path.mkdir(parents=True)
    test_bad_path.mkdir(parents=True)

    # Generate train good images
    for i in range(train_samples):
        image = _random_image()
        image_path = train_good_path / f"train_{i}.png"
        cv2.imwrite(str(image_path), image)

    # Generate val good images
    for i in range(val_samples[0]):
        image = _random_image()
        image_path = val_good_path / f"val_{i}.png"
        cv2.imwrite(str(image_path), image)
    # Generate val bad images
    for i in range(val_samples[1]):
        image = _random_image()
        image_path = val_bad_path / f"val_{i}.png"
        cv2.imwrite(str(image_path), image)

    # Generate test good images
    for i in range(test_samples[0]):
        image = _random_image()
        image_path = test_good_path / f"test_{i}.png"
        cv2.imwrite(str(image_path), image)
    # Generate test bad images
    for i in range(test_samples[1]):
        image = _random_image()
        image_path = test_bad_path / f"test_{i}.png"
        cv2.imwrite(str(image_path), image)

    return str(anomaly_dataset_path), dataset_arguments


@pytest.fixture
def anomaly_dataset(tmp_path: Path, dataset_arguments: AnomalyDatasetArguments) -> tuple[str, AnomalyDatasetArguments]:
    """Fixture used to dinamically generate anomaly dataset. By default images are random grayscales with size 10x10.

    Args:
        tmp_path: path to temporary directory
        dataset_arguments: dataset arguments

    Returns:
        path to anomaly dataset
    """
    yield _build_anomaly_dataset(tmp_path, dataset_arguments)
    if tmp_path.exists():
        shutil.rmtree(tmp_path)


@pytest.fixture(
    params=[AnomalyDatasetArguments(**{"train_samples": 10, "val_samples": (1, 1), "test_samples": (1, 1)})]
)
def base_anomaly_dataset(tmp_path: Path, request: Any) -> tuple[str, AnomalyDatasetArguments]:
    """Generate base anomaly dataset with the following parameters:
        - train_samples: 10
        - val_samples: (10, 10)
        - test_samples: (10, 10).

    Args:
        tmp_path: Path to temporary directory
        request: Pytest SubRequest object

    Yields:
        Path to anomaly dataset and dataset arguments
    """
    yield _build_anomaly_dataset(tmp_path, request.param)
    if tmp_path.exists():
        shutil.rmtree(tmp_path)



from __future__ import annotations

import json
import os

import albumentations
import pandas as pd
from torch.utils.data import DataLoader

from quadra.datamodules.base import BaseDataModule
from quadra.datasets import ImageClassificationListDataset, PatchSklearnClassificationTrainDataset
from quadra.utils.classification import find_test_image
from quadra.utils.patch.dataset import PatchDatasetInfo, load_train_file


class PatchSklearnClassificationDataModule(BaseDataModule):
    """DataModule for patch classification.

    Args:
        data_path: Location of the dataset
        name: Name of the datamodule
        train_filename: Name of the file containing the list of training samples
        exclude_filter: Filter to exclude samples from the dataset
        include_filter: Filter to include samples from the dataset
        class_to_idx: Dictionary mapping class names to indices
        seed: Random seed
        batch_size: Batch size
        num_workers: Number of workers
        train_transform: Transform to apply to the training samples
        val_transform: Transform to apply to the validation samples
        test_transform: Transform to apply to the test samples
        balance_classes: If True repeat low represented classes
        class_to_skip_training: List of classes skipped during training.
    """

    def __init__(
        self,
        data_path: str,
        class_to_idx: dict,
        name: str = "patch_classification_datamodule",
        train_filename: str = "dataset.txt",
        exclude_filter: list[str] | None = None,
        include_filter: list[str] | None = None,
        seed: int = 42,
        batch_size: int = 32,
        num_workers: int = 6,
        train_transform: albumentations.Compose | None = None,
        val_transform: albumentations.Compose | None = None,
        test_transform: albumentations.Compose | None = None,
        balance_classes: bool = False,
        class_to_skip_training: list | None = None,
        **kwargs,
    ):
        super().__init__(
            data_path=data_path,
            name=name,
            seed=seed,
            num_workers=num_workers,
            batch_size=batch_size,
            train_transform=train_transform,
            val_transform=val_transform,
            test_transform=test_transform,
            **kwargs,
        )
        self.class_to_idx = class_to_idx
        self.balance_classes = balance_classes
        self.train_filename = train_filename
        self.include_filter = include_filter
        self.exclude_filter = exclude_filter
        self.class_to_skip_training = class_to_skip_training

        self.train_folder = os.path.join(self.data_path, "train")
        self.val_folder = os.path.join(self.data_path, "val")
        self.test_folder = os.path.join(self.data_path, "test")
        self.info: PatchDatasetInfo
        self.train_dataset: PatchSklearnClassificationTrainDataset
        self.val_dataset: ImageClassificationListDataset
        self.test_dataset: ImageClassificationListDataset

    def _prepare_data(self):
        """Prepare data function."""
        if os.path.isfile(os.path.join(self.data_path, "info.json")):
            with open(os.path.join(self.data_path, "info.json")) as f:
                self.info = PatchDatasetInfo(**json.load(f))
        else:
            raise FileNotFoundError("No `info.json` file found in the dataset folder")

        split_df_list: list[pd.DataFrame] = []
        if os.path.isfile(os.path.join(self.train_folder, self.train_filename)):
            train_samples, train_labels = load_train_file(
                train_file_path=os.path.join(self.train_folder, self.train_filename),
                include_filter=self.include_filter,
                exclude_filter=self.exclude_filter,
                class_to_skip=self.class_to_skip_training,
            )
            train_df = pd.DataFrame({"samples": train_samples, "targets": train_labels})
            train_df["split"] = "train"
            split_df_list.append(train_df)
        if os.path.isdir(self.val_folder):
            val_samples, val_labels = find_test_image(
                folder=self.val_folder,
                exclude_filter=self.exclude_filter,
                include_filter=self.include_filter,
                include_none_class=False,
            )
            val_df = pd.DataFrame({"samples": val_samples, "targets": val_labels})
            val_df["split"] = "val"
            split_df_list.append(val_df)
        if os.path.isdir(self.test_folder):
            test_samples, test_labels = find_test_image(
                folder=self.test_folder,
                exclude_filter=self.exclude_filter,
                include_filter=self.include_filter,
                include_none_class=True,
            )
            test_df = pd.DataFrame({"samples": test_samples, "targets": test_labels})
            test_df["split"] = "test"
            split_df_list.append(test_df)
        if len(split_df_list) == 0:
            raise ValueError("No data found in all split folders")
        self.data = pd.concat(split_df_list, axis=0)

    def setup(self, stage: str | None = None) -> None:
        """Setup function."""
        if stage == "fit":
            self.train_dataset = PatchSklearnClassificationTrainDataset(
                data_path=self.data_path,
                class_to_idx=self.class_to_idx,
                samples=self.data[self.data["split"] == "train"]["samples"].tolist(),
                targets=self.data[self.data["split"] == "train"]["targets"].tolist(),
                transform=self.train_transform,
                balance_classes=self.balance_classes,
            )

            self.val_dataset = ImageClassificationListDataset(
                class_to_idx=self.class_to_idx,
                samples=self.data[self.data["split"] == "val"]["samples"].tolist(),
                targets=self.data[self.data["split"] == "val"]["targets"].tolist(),
                transform=self.val_transform,
                allow_missing_label=False,
            )

        elif stage in ["test", "predict"]:
            self.test_dataset = ImageClassificationListDataset(
                class_to_idx=self.class_to_idx,
                samples=self.data[self.data["split"] == "test"]["samples"].tolist(),
                targets=self.data[self.data["split"] == "test"]["targets"].tolist(),
                transform=self.test_transform,
                allow_missing_label=True,
            )

    def train_dataloader(self) -> DataLoader:
        """Return the train dataloader."""
        if not self.train_dataset_available:
            raise ValueError("No training sample is available")
        return DataLoader(
            self.train_dataset,
            batch_size=self.batch_size,
            shuffle=True,
            num_workers=self.num_workers,
            drop_last=False,
            pin_memory=True,
        )

    def val_dataloader(self) -> DataLoader:
        """Return the validation dataloader."""
        if not self.val_dataset_available:
            raise ValueError("No validation dataset is available")
        return DataLoader(
            self.val_dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.num_workers,
            drop_last=False,
            pin_memory=True,
        )

    def test_dataloader(self) -> DataLoader:
        """Return the test dataloader."""
        if not self.test_dataset_available:
            raise ValueError("No test dataset is available")

        return DataLoader(
            self.test_dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.num_workers,
            drop_last=False,
            pin_memory=True,
        )



# Original Author: jacobgil
# Title: Explainability for Vision Transformers
# Source: https://github.com/jacobgil/vit-explain (MIT license)
# Description: This is a heavily modified version of the original jacobgil code (the underlying math is still the same).
from __future__ import annotations

import math

import numpy as np
import torch
from sklearn.linear_model._base import LinearClassifierMixin


def rollout(
    attentions: list[torch.Tensor], discard_ratio: float = 0.9, head_fusion: str = "mean", aspect_ratio: float = 1.0
) -> np.ndarray:
    """Apply rollout on Attention matrices.

    Args:
        attentions: List of Attention matrices coming from different blocks
        discard_ratio: Percentage of elements to discard
        head_fusion: Strategy of fusion of attention heads
        aspect_ratio: Model inputs' width divided by height

    Returns:
        mask: Output mask, still needs a resize
    """
    result = torch.eye(attentions[0].size(-1))
    with torch.no_grad():
        for attention in attentions:
            if head_fusion == "mean":
                attention_heads_fused = attention.mean(dim=1)
            elif head_fusion == "max":
                attention_heads_fused = attention.max(dim=1)[0]
            elif head_fusion == "min":
                attention_heads_fused = attention.min(dim=1)[0]
            else:
                raise ValueError("Attention head fusion type Not supported")
            # Drop the lowest attentions, but
            # don't drop the class token
            flat = attention_heads_fused.view(attention_heads_fused.size(0), -1)
            _, indices = flat.topk(int(flat.size(-1) * discard_ratio), -1, False)
            flat.scatter_(-1, indices, 0)
            identity_matrix = torch.eye(attention_heads_fused.size(-1))
            a = (attention_heads_fused + 1.0 * identity_matrix) / 2
            a = a / a.sum(dim=-1).unsqueeze(1)
            result = torch.matmul(a, result)
    # Look at the total attention between the class token and the image patches
    mask = result[:, 0, 1:]
    batch_size = mask.size(0)
    # TODO: Non squared input-size handling can be improved. Not easy though
    height = math.floor((mask.size(-1) / aspect_ratio) ** 0.5)
    total_size = mask.size(-1)
    width = math.floor(total_size / height)
    if mask.size(-1) > (height * width):
        to_remove = mask.size(-1) - (height * width)
        mask = mask[:, :-to_remove]
    mask = mask.reshape(batch_size, height, width).numpy()
    mask = mask / mask.max(axis=(1, 2), keepdims=True)

    return mask


class VitAttentionRollout:
    """Attention gradient rollout class. Constructor registers hooks to the model's specified layers.
    Only 4 layers by default given the high load on gpu. Best gradcams obtained using all blocks.

    Args:
        model: Model
        attention_layer_names: On which layers to register the hook
        head_fusion: Strategy of fusion for attention heads
        discard_ratio: Percentage of elements to discard
    """

    def __init__(
        self,
        model: torch.nn.Module,
        attention_layer_names: list[str] | None = None,
        head_fusion: str = "mean",
        discard_ratio: float = 0.9,
    ):
        if attention_layer_names is None:
            attention_layer_names = [
                "blocks.6.attn.attn_drop",
                "blocks.7.attn.attn_drop",
                "blocks.10.attn.attn_drop",
                "blocks.11.attn.attn_drop",
            ]
        self.model = model
        self.head_fusion = head_fusion
        self.discard_ratio = discard_ratio
        self.f_hook_handles: list[torch.utils.hooks.RemovableHandle] = []
        for name, module in self.model.named_modules():
            for layer_name in attention_layer_names:
                if layer_name in name:
                    self.f_hook_handles.append(module.register_forward_hook(self.get_attention))
        self.attentions: list[torch.Tensor] = []

    # pylint: disable=unused-argument
    def get_attention(
        self,
        module: torch.nn.Module,
        inpt: torch.Tensor,
        out: torch.Tensor,
    ) -> None:
        """Hook to return attention.

        Args:
            module: Torch module
            inpt: Input tensor
            out: Output tensor, in this case the attention
        """
        self.attentions.append(out.detach().clone().cpu())

    def __call__(self, input_tensor: torch.Tensor) -> np.ndarray:
        """Called when the class instance is used as a function.

        Args:
            input_tensor: Input tensor

        Returns:
            out: Batch of output masks
        """
        self.attentions.clear()
        with torch.no_grad():
            self.model(input_tensor)
        out = rollout(
            self.attentions,
            self.discard_ratio,
            self.head_fusion,
            aspect_ratio=(input_tensor.shape[-1] / input_tensor.shape[-2]),
        )

        return out


############################################### GRAD ROLLOUT ##########################################################


def grad_rollout(
    attentions: list[torch.Tensor], gradients: list[torch.Tensor], discard_ratio: float = 0.9, aspect_ratio: float = 1.0
) -> np.ndarray:
    """Apply gradient rollout on Attention matrices.

    Args:
        attentions: Attention matrices
        gradients: Target class gradient matrices
        discard_ratio: Percentage of elements to discard
        aspect_ratio: Model inputs' width divided by height

    Returns:
        mask: Output mask, still needs a resize
    """
    result = torch.eye(attentions[0].size(-1))
    with torch.no_grad():
        for attention, grad in zip(attentions, gradients, strict=False):
            weights = grad
            attention_heads_fused = torch.mean((attention * weights), dim=1)
            attention_heads_fused[attention_heads_fused < 0] = 0
            # Drop the lowest attentions, but
            # don't drop the class token
            flat = attention_heads_fused.view(attention_heads_fused.size(0), -1)
            _, indices = flat.topk(int(flat.size(-1) * discard_ratio), -1, False)
            flat.scatter_(-1, indices, 0)
            I = torch.eye(attention_heads_fused.size(-1))
            a = (attention_heads_fused + 1.0 * I) / 2
            a = a / a.sum(dim=-1).unsqueeze(1)
            result = torch.matmul(a, result)
    # Look at the total attention between the class token,
    # and the image patches
    mask = result[:, 0, 1:]
    batch_size = mask.size(0)
    # TODO: Non squared input-size handling can be improved. Not easy though
    height = math.floor((mask.size(-1) / aspect_ratio) ** 0.5)
    total_size = mask.size(-1)
    width = math.floor(total_size / height)
    if mask.size(-1) > (height * width):
        to_remove = mask.size(-1) - (height * width)
        mask = mask[:, :-to_remove]
    mask = mask.reshape(batch_size, height, width).numpy()
    mask = mask / mask.max(axis=(1, 2), keepdims=True)

    return mask


class VitAttentionGradRollout:
    """Attention gradient rollout class. Constructor registers hooks to the model's specified layers.
    Only 4 layers by default given the high load on gpu. Best gradcams obtained using all blocks.

    Args:
        model: Pytorch model
        attention_layer_names: On which layers to register the hooks
        discard_ratio: Percentage of elements to discard
        classifier: Scikit-learn classifier. Leave it to None if model already has a classifier on top.
    """

    def __init__(  # pylint: disable=W0102
        self,
        model: torch.nn.Module,
        attention_layer_names: list[str] | None = None,
        discard_ratio: float = 0.9,
        classifier: LinearClassifierMixin | None = None,
        example_input: torch.Tensor | None = None,
    ):
        if attention_layer_names is None:
            attention_layer_names = [
                "blocks.6.attn.attn_drop",
                "blocks.7.attn.attn_drop",
                "blocks.10.attn.attn_drop",
                "blocks.11.attn.attn_drop",
            ]

        if classifier is not None:
            if example_input is None:
                raise ValueError(
                    "Must provide an input example to LinearModelPytorchWrapper when classifier is not None"
                )
            self.model = LinearModelPytorchWrapper(
                backbone=model,
                linear_classifier=classifier,
                example_input=example_input,
                device=next(model.parameters()).device,
            )
        else:
            self.model = model  # type: ignore[assignment]

        self.discard_ratio = discard_ratio
        self.f_hook_handles: list[torch.utils.hooks.RemovableHandle] = []
        self.b_hook_handles: list[torch.utils.hooks.RemovableHandle] = []
        for name, module in self.model.named_modules():
            for layer_name in attention_layer_names:
                if layer_name in name:
                    self.f_hook_handles.append(module.register_forward_hook(self.get_attention))
                    self.b_hook_handles.append(module.register_backward_hook(self.get_attention_gradient))
        self.attentions: list[torch.Tensor] = []
        self.attention_gradients: list[torch.Tensor] = []
        # Activate gradients
        blocks_list = [x.split("blocks")[1].split(".attn")[0] for x in attention_layer_names]
        for name, module in model.named_modules():
            for p in module.parameters():
                if "blocks" in name and any(x in name for x in blocks_list):
                    p.requires_grad = True

    # pylint: disable=unused-argument
    def get_attention(
        self,
        module: torch.nn.Module,
        inpt: torch.Tensor,
        out: torch.Tensor,
    ) -> None:
        """Hook to return attention.

        Args:
            module: Torch module
            inpt: Input tensor
            out: Output tensor, in this case the attention
        """
        self.attentions.append(out.detach().clone().cpu())

    # pylint: disable=unused-argument
    def get_attention_gradient(
        self,
        module: torch.nn.Module,
        grad_input: torch.Tensor,
        grad_output: torch.Tensor,
    ) -> None:
        """Hook to return attention.

        Args:
            module: Torch module
            grad_input: Gradients' input tensor
            grad_output: Gradients' output tensor, in this case the attention
        """
        self.attention_gradients.append(grad_input[0].detach().clone().cpu())

    def __call__(self, input_tensor: torch.Tensor, targets_list: list[int]) -> np.ndarray:
        """Called when the class instance is used as a function.

        Args:
            input_tensor: Model's input tensor
            targets_list: List of targets. If None, argmax is used

        Returns:
            out: Batch of output masks
        """
        self.attentions.clear()
        self.attention_gradients.clear()

        self.model.zero_grad(set_to_none=True)
        self.model.to(input_tensor.device)
        output = self.model(input_tensor).cpu()

        class_mask = torch.zeros(output.shape)
        if targets_list is None:
            targets_list = output.argmax(dim=1)
        class_mask[torch.arange(output.shape[0]), targets_list] = 1
        loss = (output * class_mask).sum()
        loss.backward()
        out = grad_rollout(
            self.attentions,
            self.attention_gradients,
            self.discard_ratio,
            aspect_ratio=(input_tensor.shape[-1] / input_tensor.shape[-2]),
        )

        return out


class LinearModelPytorchWrapper(torch.nn.Module):
    """Pytorch wrapper for scikit-learn linear models.

    Args:
        backbone: Backbone
        linear_classifier: ScikitLearn linear classifier model
        example_input: Input example needed to obtain output shape
        device: The device to use. Defaults to "cpu"
    """

    def __init__(
        self,
        backbone: torch.nn.Module,
        linear_classifier: LinearClassifierMixin,
        example_input: torch.Tensor,
        device: torch.device,
    ):
        super().__init__()
        self.device = device
        self.backbone = backbone.to(device)
        if not isinstance(linear_classifier, LinearClassifierMixin):
            raise TypeError("Classifier is not of type LinearClassifierMixin.")
        self.num_classes = len(linear_classifier.classes_)
        self.linear_classifier = linear_classifier
        with torch.no_grad():
            output = self.backbone(example_input.to(device))
            num_filters = output.shape[-1]

        self.classifier = torch.nn.Linear(num_filters, self.num_classes).to(device)
        self.classifier.weight.data = torch.from_numpy(linear_classifier.coef_).float()
        self.classifier.bias.data = torch.from_numpy(linear_classifier.intercept_).float()

    def forward(self, x):
        if self.num_classes == 2:
            class_one_probabilities = torch.nn.Sigmoid()(self.classifier(self.backbone(x)))
            class_zero_probabilities = 1 - class_one_probabilities
            two_class_probs = torch.cat((class_zero_probabilities, class_one_probabilities), dim=1)

            return two_class_probs

        return torch.nn.Softmax(dim=1)(self.classifier(self.backbone(x)))



from __future__ import annotations

import glob
import os
import random
import shutil
from dataclasses import dataclass
from pathlib import Path
from typing import Any

import cv2
import numpy as np
import pytest

from quadra.utils.patch import generate_patch_dataset, get_image_mask_association
from quadra.utils.tests.helpers import _random_image


@dataclass
class ClassificationDatasetArguments:
    """Classification dataset arguments.

    Args:
        samples: number of samples per class
        classes: class names, if set it must be the same length as samples
        val_size: validation set size
        test_size: test set size
    """

    samples: list[int]
    classes: list[str] | None = None
    val_size: float | None = None
    test_size: float | None = None


@dataclass
class ClassificationMultilabelDatasetArguments:
    """Classification dataset arguments.

    Args:
        samples: number of samples per class
        classes: class names, if set it must be the same length as samples
        val_size: validation set size
        test_size: test set size
        percentage_other_classes: probability of adding other classes to the labels of each sample
    """

    samples: list[int]
    classes: list[str] | None = None
    val_size: float | None = None
    test_size: float | None = None
    percentage_other_classes: float | None = 0.0


@dataclass
class ClassificationPatchDatasetArguments:
    """Classification patch dataset arguments.

    Args:
        samples: number of samples per class
        overlap: overlap between patches
        patch_size: patch size
        patch_number: number of patches
        classes: class names, if set it must be the same length as samples
        val_size: validation set size
        test_size: test set size
        annotated_good: list of class names that are considered as good annotations (E.g. ["good"])
    """

    samples: list[int]
    overlap: float
    patch_size: tuple[int, int] | None = None
    patch_number: tuple[int, int] | None = None
    classes: list[str] | None = None
    val_size: float | None = 0.0
    test_size: float | None = 0.0
    annotated_good: list[str] | None = None


def _build_classification_dataset(
    tmp_path: Path, dataset_arguments: ClassificationDatasetArguments
) -> tuple[str, ClassificationDatasetArguments]:
    """Generate classification dataset. If val_size or test_size are set, it will generate a train.txt, val.txt and
        test.txt file in the dataset directory. By default generated images are 10x10 pixels.

    Args:
        tmp_path: path to temporary directory
        dataset_arguments: dataset arguments

    Returns:
        Tuple containing path to created dataset and dataset arguments
    """
    classification_dataset_path = tmp_path / "classification_dataset"
    classification_dataset_path.mkdir()

    classes = dataset_arguments.classes if dataset_arguments.classes else range(len(dataset_arguments.samples))

    for class_name, samples in zip(classes, dataset_arguments.samples):
        class_path = classification_dataset_path / str(class_name)
        class_path.mkdir()
        for i in range(samples):
            image = _random_image()
            image_path = class_path / f"{class_name}_{i}.png"
            cv2.imwrite(str(image_path), image)

    if dataset_arguments.val_size is not None or dataset_arguments.test_size is not None:
        all_images = glob.glob(os.path.join(str(classification_dataset_path), "**", "*.png"))
        all_images = [f"{os.path.basename(os.path.dirname(image))}/{os.path.basename(image)}" for image in all_images]
        val_size = dataset_arguments.val_size if dataset_arguments.val_size is not None else 0
        test_size = dataset_arguments.test_size if dataset_arguments.test_size is not None else 0
        train_size = 1 - val_size - test_size

        # pylint: disable=unbalanced-tuple-unpacking
        train_images, val_images, test_images = np.split(
            np.random.permutation(all_images),
            [int(train_size * len(all_images)), int((train_size + val_size) * len(all_images))],
        )

        with open(classification_dataset_path / "train.txt", "w") as f:
            f.write("\n".join(train_images))

        with open(classification_dataset_path / "val.txt", "w") as f:
            f.write("\n".join(val_images))

        with open(classification_dataset_path / "test.txt", "w") as f:
            f.write("\n".join(test_images))

    return str(classification_dataset_path), dataset_arguments


@pytest.fixture
def classification_dataset(
    tmp_path: Path, dataset_arguments: ClassificationDatasetArguments
) -> tuple[str, ClassificationDatasetArguments]:
    """Generate classification dataset. If val_size or test_size are set, it will generate a train.txt, val.txt and
        test.txt file in the dataset directory. By default generated images are 10x10 pixels.

    Args:
        tmp_path: path to temporary directory
        dataset_arguments: dataset arguments

    Yields:
        Tuple containing path to created dataset and dataset arguments
    """
    yield _build_classification_dataset(tmp_path, dataset_arguments)
    if tmp_path.exists():
        shutil.rmtree(tmp_path)


@pytest.fixture(
    params=[
        ClassificationDatasetArguments(
            **{"samples": [10, 10], "classes": ["class_1", "class_2"], "val_size": 0.1, "test_size": 0.1}
        )
    ]
)
def base_classification_dataset(tmp_path: Path, request: Any) -> tuple[str, ClassificationDatasetArguments]:
    """Generate base classification dataset with the following parameters:
        - 10 samples per class
        - 2 classes (class_1 and class_2)
        By default generated images are grayscale and 10x10 pixels.

    Args:
        tmp_path: path to temporary directory
        request: pytest request

    Yields:
        Tuple containing path to created dataset and dataset arguments
    """
    yield _build_classification_dataset(tmp_path, request.param)
    if tmp_path.exists():
        shutil.rmtree(tmp_path)


def _build_multilabel_classification_dataset(
    tmp_path: Path, dataset_arguments: ClassificationMultilabelDatasetArguments
) -> tuple[str, ClassificationMultilabelDatasetArguments]:
    """Generate a multilabel classification dataset.
        Generates a samples.txt file in the dataset directory containing the path to the image and the corresponding
        classes. If val_size or test_size are set, it will generate a train.txt, val.txt and test.txt file in the
        dataset directory. By default generated images are 10x10 pixels.

    Args:
        tmp_path: path to temporary directory
        dataset_arguments: dataset arguments

    Returns:
        Tuple containing path to created dataset and dataset arguments
    """
    classification_dataset_path = tmp_path / "multilabel_classification_dataset"
    images_path = classification_dataset_path / "images"
    classification_dataset_path.mkdir()
    images_path.mkdir()

    classes = dataset_arguments.classes if dataset_arguments.classes else range(len(dataset_arguments.samples))
    percentage_other_classes = dataset_arguments.percentage_other_classes

    generated_samples = []
    counter = 0
    for class_name, samples in zip(classes, dataset_arguments.samples):
        for _ in range(samples):
            image = _random_image()
            image_path = images_path / f"{counter}.png"
            counter += 1
            cv2.imwrite(str(image_path), image)
            targets = [class_name]
            targets = targets + [
                cl_name for cl_name in classes if cl_name != class_name and random.random() < percentage_other_classes
            ]
            generated_samples.append(f"images/{image_path.name},{','.join(targets)}")

    with open(classification_dataset_path / "samples.txt", "w") as f:
        f.write("\n".join(generated_samples))

    if dataset_arguments.val_size is not None or dataset_arguments.test_size is not None:
        val_size = dataset_arguments.val_size if dataset_arguments.val_size is not None else 0
        test_size = dataset_arguments.test_size if dataset_arguments.test_size is not None else 0
        train_size = 1 - val_size - test_size

        # pylint: disable=unbalanced-tuple-unpacking
        train_images, val_images, test_images = np.split(
            np.random.permutation(generated_samples),
            [int(train_size * len(generated_samples)), int((train_size + val_size) * len(generated_samples))],
        )

        with open(classification_dataset_path / "train.txt", "w") as f:
            f.write("\n".join(train_images))

        with open(classification_dataset_path / "val.txt", "w") as f:
            f.write("\n".join(val_images))

        with open(classification_dataset_path / "test.txt", "w") as f:
            f.write("\n".join(test_images))

    return str(classification_dataset_path), dataset_arguments


@pytest.fixture
def multilabel_classification_dataset(
    tmp_path: Path, dataset_arguments: ClassificationMultilabelDatasetArguments
) -> tuple[str, ClassificationMultilabelDatasetArguments]:
    """Fixture to dinamically generate a multilabel classification dataset.
        Generates a samples.txt file in the dataset directory containing the path to the image and the corresponding
        classes. If val_size or test_size are set, it will generate a train.txt, val.txt and test.txt file in the
        dataset directory. By default generated images are 10x10 pixels.

    Args:
        tmp_path: path to temporary directory
        dataset_arguments: dataset arguments

    Returns:
        Tuple containing path to created dataset and dataset arguments
    """
    yield _build_multilabel_classification_dataset(tmp_path, dataset_arguments)
    if tmp_path.exists():
        shutil.rmtree(tmp_path)


@pytest.fixture(
    params=[
        ClassificationMultilabelDatasetArguments(
            **{
                "samples": [10, 10, 10],
                "classes": ["class_1", "class_2", "class_3"],
                "val_size": 0.1,
                "test_size": 0.1,
                "percentage_other_classes": 0.3,
            }
        )
    ]
)
def base_multilabel_classification_dataset(
    tmp_path: Path, request: Any
) -> tuple[str, ClassificationMultilabelDatasetArguments]:
    """Fixture to generate base multilabel classification dataset with the following parameters:
        - 10 samples per class
        - 3 classes (class_1, class_2 and class_3)
        - 10% of samples in validation set
        - 10% of samples in test set
        - 30% of possibility to add each other class to the sample
        By default generated images are grayscale and 10x10 pixels.

    Args:
        tmp_path: path to temporary directory
        request: pytest request

    Yields:
        Tuple containing path to created dataset and dataset arguments
    """
    yield _build_multilabel_classification_dataset(tmp_path, request.param)
    if tmp_path.exists():
        shutil.rmtree(tmp_path)


def _build_classification_patch_dataset(
    tmp_path: Path, dataset_arguments: ClassificationDatasetArguments
) -> tuple[str, ClassificationDatasetArguments, dict[str, int]]:
    """Generate a classification patch dataset. By default generated images are 224x224 pixels
        and associated masks contains a 50x50 pixels square with the corresponding image class, so at the current stage
        is not possible to have images with multiple annotations. The patch dataset will be generated using the standard
        parameters of generate_patch_dataset function.

    Args:
        tmp_path: path to temporary directory
        dataset_arguments: dataset arguments

    Returns:
        Tuple containing path to created dataset, dataset arguments and class to index mapping
    """
    initial_dataset_path = tmp_path / "initial_dataset"
    initial_dataset_path.mkdir()

    images_path = initial_dataset_path / "images"
    masks_path = initial_dataset_path / "masks"
    images_path.mkdir()
    masks_path.mkdir()

    classes = dataset_arguments.classes if dataset_arguments.classes else range(len(dataset_arguments.samples))

    class_to_idx = {class_name: i for i, class_name in enumerate(classes)}

    for class_name, samples in zip(classes, dataset_arguments.samples):
        for i in range(samples):
            image = _random_image(size=(224, 224))
            mask = np.zeros((224, 224), dtype=np.uint8)
            mask[100:150, 100:150] = class_to_idx[class_name]
            image_path = images_path / f"{class_name}_{i}.png"
            mask_path = masks_path / f"{class_name}_{i}.png"
            cv2.imwrite(str(image_path), image)
            cv2.imwrite(str(mask_path), mask)

    patch_dataset_path = tmp_path / "patch_dataset"
    patch_dataset_path.mkdir()

    data_dictionary = get_image_mask_association(data_folder=str(images_path), mask_folder=str(masks_path))

    _ = generate_patch_dataset(
        data_dictionary=data_dictionary,
        class_to_idx=class_to_idx,
        val_size=dataset_arguments.val_size,
        test_size=dataset_arguments.test_size,
        patch_number=dataset_arguments.patch_number,
        patch_size=dataset_arguments.patch_size,
        overlap=dataset_arguments.overlap,
        output_folder=str(patch_dataset_path),
        annotated_good=dataset_arguments.annotated_good,
    )

    return str(patch_dataset_path), dataset_arguments, class_to_idx


@pytest.fixture
def classification_patch_dataset(
    tmp_path: Path, dataset_arguments: ClassificationDatasetArguments
) -> tuple[str, ClassificationDatasetArguments, dict[str, int]]:
    """Fixture to dinamically generate a classification patch dataset.

        By default generated images are 224x224 pixels
        and associated masks contains a 50x50 pixels square with the corresponding image class, so at the current stage
        is not possible to have images with multiple annotations. The patch dataset will be generated using the standard
        parameters of generate_patch_dataset function.

    Args:
        tmp_path: path to temporary directory
        dataset_arguments: dataset arguments

    Yields:
        Tuple containing path to created dataset, dataset arguments and class to index mapping
    """
    yield _build_classification_patch_dataset(tmp_path, dataset_arguments)
    if tmp_path.exists():
        shutil.rmtree(tmp_path)


@pytest.fixture(
    params=[
        ClassificationPatchDatasetArguments(
            **{
                "samples": [5, 5, 5],
                "classes": ["bg", "a", "b"],
                "patch_number": [2, 2],
                "overlap": 0,
                "val_size": 0.1,
                "test_size": 0.1,
            }
        )
    ]
)
def base_patch_classification_dataset(
    tmp_path: Path, request: Any
) -> tuple[str, ClassificationDatasetArguments, dict[str, int]]:
    """Generate a classification patch dataset with the following parameters:
        - 3 classes named bg, a and b
        - 5, 5 and 5 samples for each class
        - 2 horizontal patches and 2 vertical patches
        - 0% overlap
        - 10% validation set
        - 10% test set.

    Args:
        tmp_path: path to temporary directory
        request: pytest SubRequest object
    """
    yield _build_classification_patch_dataset(tmp_path, request.param)
    if tmp_path.exists():
        shutil.rmtree(tmp_path)



from __future__ import annotations

import math
from collections.abc import Callable, Sized
from typing import Any

import sklearn
import torch
from pytorch_lightning.core.optimizer import LightningOptimizer
from torch import nn
from torch.optim import Optimizer

from quadra.modules.base import SSLModule


class BYOL(SSLModule):
    """BYOL module, inspired by https://arxiv.org/abs/2006.07733.

    Args:
        student : student model.
        teacher : teacher model.
        student_projection_mlp : student projection MLP.
        student_prediction_mlp : student prediction MLP.
        teacher_projection_mlp : teacher projection MLP.
        criterion : loss function.
        classifier: Standard sklearn classifier.
        optimizer: optimizer of the training. If None a default Adam is used.
        lr_scheduler: lr scheduler. If None a default ReduceLROnPlateau is used.
        lr_scheduler_interval: interval at which the lr scheduler is updated.
        teacher_momentum: momentum of the teacher parameters.
        teacher_momentum_cosine_decay: whether to use cosine decay for the teacher momentum. Default: True
    """

    def __init__(
        self,
        student: nn.Module,
        teacher: nn.Module,
        student_projection_mlp: nn.Module,
        student_prediction_mlp: nn.Module,
        teacher_projection_mlp: nn.Module,
        criterion: nn.Module,
        classifier: sklearn.base.ClassifierMixin | None = None,
        optimizer: Optimizer | None = None,
        lr_scheduler: object | None = None,
        lr_scheduler_interval: str | None = "epoch",
        teacher_momentum: float = 0.9995,
        teacher_momentum_cosine_decay: bool | None = True,
    ):
        super().__init__(
            model=student,
            criterion=criterion,
            classifier=classifier,
            optimizer=optimizer,
            lr_scheduler=lr_scheduler,
            lr_scheduler_interval=lr_scheduler_interval,
        )
        # Student model
        self.max_steps: int
        self.student_projection_mlp = student_projection_mlp
        self.student_prediction_mlp = student_prediction_mlp

        # Teacher model
        self.teacher = teacher
        self.teacher_projection_mlp = teacher_projection_mlp
        self.teacher_initialized = False
        self.teacher_momentum = teacher_momentum
        self.teacher_momentum_cosine_decay = teacher_momentum_cosine_decay

        self.initialize_teacher()

    def initialize_teacher(self):
        """Initialize teacher from the state dict of the student one,
        checking also that student model requires greadient correctly.
        """
        self.teacher_projection_mlp.load_state_dict(self.student_projection_mlp.state_dict())
        for p in self.teacher_projection_mlp.parameters():
            p.requires_grad = False

        self.teacher.load_state_dict(self.model.state_dict())
        for p in self.teacher.parameters():
            p.requires_grad = False

        for p in self.student_projection_mlp.parameters():
            assert p.requires_grad is True
        for p in self.student_prediction_mlp.parameters():
            assert p.requires_grad is True

        self.teacher_initialized = True

    def update_teacher(self):
        """Update teacher given `self.teacher_momentum` by an exponential moving average
        of the student parameters, that is: theta_t * tau + theta_s * (1 - tau), where
        `theta_{s,t}` are the parameters of the student and the teacher model, while `tau` is the
        teacher momentum. If `self.teacher_momentum_cosine_decay` is True, then the teacher
        momentum will follow a cosine scheduling from `self.teacher_momentum` to 1:
        tau = 1 - (1 - tau) * (cos(pi * t / T) + 1) / 2, where `t` is the current step and
        `T` is the max number of steps.
        """
        with torch.no_grad():
            if self.teacher_momentum_cosine_decay:
                teacher_momentum = (
                    1
                    - (1 - self.teacher_momentum)
                    * (math.cos(math.pi * self.trainer.global_step / self.max_steps) + 1)
                    / 2
                )
            else:
                teacher_momentum = self.teacher_momentum
            self.log("teacher_momentum", teacher_momentum, prog_bar=True)
            for student_ps, teacher_ps in zip(
                list(self.model.parameters()) + list(self.student_projection_mlp.parameters()),
                list(self.teacher.parameters()) + list(self.teacher_projection_mlp.parameters()),
                strict=False,
            ):
                teacher_ps.data = teacher_ps.data * teacher_momentum + (1 - teacher_momentum) * student_ps.data

    def on_train_start(self) -> None:
        if isinstance(self.trainer.train_dataloader, Sized) and isinstance(self.trainer.max_epochs, int):
            self.max_steps = len(self.trainer.train_dataloader) * self.trainer.max_epochs
        else:
            raise ValueError("BYOL requires `max_epochs` to be set and `train_dataloader` to be initialized.")

    def training_step(self, batch: tuple[list[torch.Tensor], torch.Tensor], *args: Any) -> torch.Tensor:
        [image1, image2], _ = batch

        online_pred_one = self.student_prediction_mlp(self.student_projection_mlp(self.model(image1)))
        online_pred_two = self.student_prediction_mlp(self.student_projection_mlp(self.model(image2)))

        with torch.no_grad():
            target_proj_one = self.teacher_projection_mlp(self.teacher(image1))
            target_proj_two = self.teacher_projection_mlp(self.teacher(image2))

        loss_one = self.criterion(online_pred_one, target_proj_two.detach())
        loss_two = self.criterion(online_pred_two, target_proj_one.detach())
        loss = loss_one + loss_two

        self.log(name="loss", value=loss, on_step=True, on_epoch=True, prog_bar=True)
        return loss

    def optimizer_step(
        self,
        epoch: int,
        batch_idx: int,
        optimizer: Optimizer | LightningOptimizer,
        optimizer_closure: Callable[[], Any] | None = None,
    ) -> None:
        """Override optimizer step to update the teacher parameters."""
        super().optimizer_step(
            epoch,
            batch_idx,
            optimizer,
            optimizer_closure=optimizer_closure,
        )
        self.update_teacher()

    def calculate_accuracy(self, batch):
        """Calculate accuracy on the given batch."""
        images, labels = batch
        embedding = self.model(images).detach().cpu().numpy()
        predictions = self.classifier.predict(embedding)
        labels = labels.detach()
        acc = self.val_acc(torch.tensor(predictions, device=self.device), labels)

        return acc

    def on_test_epoch_start(self) -> None:
        self.fit_estimator()

    def test_step(self, batch, *args: list[Any]) -> None:
        """Calculate accuracy on the test set for the given batch."""
        acc = self.calculate_accuracy(batch)
        self.log(name="test_acc", value=acc, on_step=False, on_epoch=True, prog_bar=True)
        return acc



import torch
from torch.nn.functional import cosine_similarity


def cosine_align_loss(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """Computes mean of cosine distance based on similarity mean(1 - cosine_similarity).

    Args:
        x: feature n1
        y: feature n2.

    Returns:
        cosine align loss
    """
    cos = 1 - cosine_similarity(x, y, dim=1)
    return torch.mean(cos)


# Source: https://arxiv.org/pdf/2005.10242.pdf
def align_loss(x: torch.Tensor, y: torch.Tensor, alpha: int = 2) -> torch.Tensor:
    """Mean(l2^alpha).

    Args:
        x: feature n1
        y: feature n2
        alpha: pow of the norm loss.

    Returns:
        Align loss
    """
    norm = torch.norm(x - y, p=2, dim=1)
    return torch.mean(torch.pow(norm, alpha))


def uniform_loss(x: torch.Tensor, t: float = 2.0) -> torch.Tensor:
    """log(mean(exp(-t*dist_p2))).

    Args:
        x: feature tensor
        t: temperature of the dist_p2.

    Returns:
        Uniform loss
    """
    return torch.log(torch.mean(torch.exp(torch.pow(torch.pdist(x, p=2), 2) * -t)))



import os
from pathlib import Path

from torchvision.datasets.utils import download_and_extract_archive

from quadra.datamodules import AnomalyDataModule
from quadra.utils.utils import get_logger

log = get_logger(__name__)


DATASET_BASE_URL = "https://www.mydrive.ch/shares/38536/3830184030e49fe74747669442f0f282/download/"

DATASET_URL = {
    "bottle": DATASET_BASE_URL + "420937370-1629951468/bottle.tar.xz",
    "capsule": DATASET_BASE_URL + "420937454-1629951595/capsule.tar.xz",
    "carpet": DATASET_BASE_URL + "420937484-1629951672/carpet.tar.xz",
    "grid": DATASET_BASE_URL + "420937487-1629951814/grid.tar.xz",
    "hazelnut": DATASET_BASE_URL + "420937545-1629951845/hazelnut.tar.xz",
    "leather": DATASET_BASE_URL + "420937607-1629951964/leather.tar.xz",
    "metal_nut": DATASET_BASE_URL + "420937637-1629952063/metal_nut.tar.xz",
    "pill": DATASET_BASE_URL + "420938129-1629953099/pill.tar.xz",
    "screw": DATASET_BASE_URL + "420938130-1629953152/screw.tar.xz",
    "tile": DATASET_BASE_URL + "420938133-1629953189/tile.tar.xz",
    "toothbrush": DATASET_BASE_URL + "420938134-1629953256/toothbrush.tar.xz",
    "transistor": DATASET_BASE_URL + "420938166-1629953277/transistor.tar.xz",
    "wood": DATASET_BASE_URL + "420938383-1629953354/wood.tar.xz",
    "zipper": DATASET_BASE_URL + "420938385-1629953449/zipper.tar.xz",
}


class MVTecDataModule(AnomalyDataModule):
    """Standard anomaly datamodule with automatic download of the MVTec dataset."""

    def __init__(self, data_path: str, category: str, **kwargs):
        if category not in DATASET_URL:
            raise ValueError(f"Unknown category {category}. Available categories are {list(DATASET_URL.keys())}")

        super().__init__(data_path=data_path, category=category, **kwargs)

    def download_data(self) -> None:
        """Download the MVTec dataset."""
        if self.category is None:
            raise ValueError("Category must be specified for MVTec dataset.")

        if os.path.exists(self.data_path):
            log.info("The path %s already exists. Skipping download.", os.path.join(self.data_path, self.category))
            return

        log.info("Downloading and extracting MVTec dataset for category %s to %s", self.category, self.data_path)
        # self.data_path is the path to the category folder that will be created by the download_and_extract_archive
        data_path_no_category = str(Path(self.data_path).parent)
        download_and_extract_archive(DATASET_URL[self.category], data_path_no_category, remove_finished=True)

    def _prepare_data(self) -> None:
        """Prepare the MVTec dataset."""
        self.download_data()
        return super()._prepare_data()



import time

import hydra
import matplotlib
from omegaconf import DictConfig
from pytorch_lightning import seed_everything

from quadra.tasks.base import Task
from quadra.utils.resolver import register_resolvers
from quadra.utils.utils import get_logger, load_envs, setup_opencv
from quadra.utils.validator import validate_config

load_envs()
register_resolvers()


matplotlib.use("Agg")
log = get_logger(__name__)


@hydra.main(config_path="configs/", config_name="config.yaml", version_base="1.3.0")
def main(config: DictConfig):
    """Main entry function for any of the tasks."""
    if config.validate:
        start = time.time()
        validate_config(config)
        stop = time.time()
        log.info("Config validation took %f seconds", stop - start)

    from quadra.utils import utils  # pylint: disable=import-outside-toplevel

    utils.extras(config)

    # Prints the resolved configuration to the console
    if config.get("print_config"):
        utils.print_config(config, resolve=True)

    # Set seed for random number generators in pytorch, numpy and python.random
    seed_everything(config.core.seed, workers=True)
    setup_opencv()

    # Run specified task using the configuration composition
    task: Task = hydra.utils.instantiate(config.task, config, _recursive_=False)
    task.execute()


if __name__ == "__main__":
    # pylint: disable=no-value-for-parameter
    main()



from __future__ import annotations

from collections.abc import Callable
from typing import Any

import sklearn
import torch
from pytorch_lightning.core.optimizer import LightningOptimizer
from torch import nn
from torch.optim import Optimizer

from quadra.modules.ssl import BYOL
from quadra.utils.models import clip_gradients
from quadra.utils.utils import get_logger

log = get_logger(__name__)


class Dino(BYOL):
    """DINO pytorch-lightning module.

    Args:
        student : student model
        teacher : teacher model
        student_projection_mlp : student projection MLP
        teacher_projection_mlp : teacher projection MLP
        criterion : loss function
        freeze_last_layer : number of layers to freeze in the student model. Default: 1
        classifier: Standard sklearn classifier
        optimizer: optimizer of the training. If None a default Adam is used.
        lr_scheduler: lr scheduler. If None a default ReduceLROnPlateau is used.
        lr_scheduler_interval: interval at which the lr scheduler is updated.
        teacher_momentum: momentum of the teacher parameters
        teacher_momentum_cosine_decay: whether to use cosine decay for the teacher momentum
    """

    def __init__(
        self,
        student: nn.Module,
        teacher: nn.Module,
        student_projection_mlp: nn.Module,
        teacher_projection_mlp: nn.Module,
        criterion: nn.Module,
        freeze_last_layer: int = 1,
        classifier: sklearn.base.ClassifierMixin | None = None,
        optimizer: Optimizer | None = None,
        lr_scheduler: object | None = None,
        lr_scheduler_interval: str | None = "epoch",
        teacher_momentum: float = 0.9995,
        teacher_momentum_cosine_decay: bool | None = True,
    ):
        super().__init__(
            student=student,
            teacher=teacher,
            student_projection_mlp=student_projection_mlp,
            student_prediction_mlp=nn.Identity(),
            teacher_projection_mlp=teacher_projection_mlp,
            criterion=criterion,
            teacher_momentum=teacher_momentum,
            teacher_momentum_cosine_decay=teacher_momentum_cosine_decay,
            classifier=classifier,
            optimizer=optimizer,
            lr_scheduler=lr_scheduler,
            lr_scheduler_interval=lr_scheduler_interval,
        )
        self.freeze_last_layer = freeze_last_layer

    def initialize_teacher(self):
        """Initialize teacher from the state dict of the student one,
        checking also that student model requires greadient correctly.
        """
        self.teacher_projection_mlp.load_state_dict(self.student_projection_mlp.state_dict())
        for p in self.teacher_projection_mlp.parameters():
            p.requires_grad = False

        self.teacher.load_state_dict(self.model.state_dict())
        for p in self.teacher.parameters():
            p.requires_grad = False

        all_frozen = True
        for p in self.model.parameters():
            all_frozen = all_frozen and (not p.requires_grad)

        if all_frozen:
            log.warning(
                "All parameters of the student model are frozen, the model will not be trained, automatically"
                " unfreezing all the layers"
            )

            for p in self.model.parameters():
                p.requires_grad = True

        for name, p in self.student_projection_mlp.named_parameters():
            if name != "last_layer.weight_g":
                assert p.requires_grad is True

        self.teacher_initialized = True

    def student_multicrop_forward(self, x: list[torch.Tensor]) -> torch.Tensor:
        """Student forward on the multicrop imges.

        Args:
            x: List of torch.Tensor containing multicropped augmented images

        Returns:
            torch.Tensor: a tensor of shape NxBxD, where N is the number crops
                corresponding to the length of the input list `x`, B is the batch size
                and D is the output dimension
        """
        n_crops = len(x)
        concatenated = torch.cat(x, dim=0)  # (n_samples * n_crops, C, H, W)
        embedding = self.model(concatenated)  # (n_samples * n_crops, in_dim)
        logits = self.student_projection_mlp(embedding)  # (n_samples * n_crops, out_dim)
        chunks = logits.chunk(n_crops)  # n_crops * (n_samples, out_dim)
        return chunks

    def teacher_multicrop_forward(self, x: list[torch.Tensor]) -> torch.Tensor:
        """Teacher forward on the multicrop imges.

        Args:
            x: List of torch.Tensor containing multicropped augmented images

        Returns:
            torch.Tensor: a tensor of shape NxBxD, where N is the number crops
                corresponding to the length of the input list `x`, B is the batch size
                and D is the output dimension
        """
        n_crops = len(x)
        concatenated = torch.cat(x, dim=0)  # (n_samples * n_crops, C, H, W)
        embedding = self.teacher(concatenated)  # (n_samples * n_crops, in_dim)
        logits = self.teacher_projection_mlp(embedding)  # (n_samples * n_crops, out_dim)
        chunks = logits.chunk(n_crops)  # n_crops * (n_samples, out_dim)
        return chunks

    def cancel_gradients_last_layer(self, epoch: int, freeze_last_layer: int):
        """Zero out the gradient of the last layer, as specified in the paper.

        Args:
            epoch: current epoch
            freeze_last_layer: maximum freeze epoch: if `epoch` >= `freeze_last_layer`
                then the gradient of the last layer will not be freezed
        """
        if epoch >= freeze_last_layer:
            return
        for n, p in self.student_projection_mlp.named_parameters():
            if "last_layer" in n:
                p.grad = None

    def training_step(self, batch: tuple[list[torch.Tensor], torch.Tensor], *args: Any) -> torch.Tensor:
        images, _ = batch
        with torch.no_grad():
            teacher_output = self.teacher_multicrop_forward(images[:2])

        student_output = self.student_multicrop_forward(images)
        loss = self.criterion(self.current_epoch, student_output, teacher_output)

        self.log(name="loss", value=loss, on_step=True, on_epoch=True, prog_bar=True)
        return loss

    def configure_gradient_clipping(
        self,
        optimizer: Optimizer,
        gradient_clip_val: int | float | None = None,
        gradient_clip_algorithm: str | None = None,
    ):
        """Configure gradient clipping for the optimizer."""
        if gradient_clip_algorithm is not None and gradient_clip_val is not None:
            clip_gradients(self.model, gradient_clip_val)
            clip_gradients(self.student_projection_mlp, gradient_clip_val)
        self.cancel_gradients_last_layer(self.current_epoch, self.freeze_last_layer)

    def optimizer_step(
        self,
        epoch: int,
        batch_idx: int,
        optimizer: Optimizer | LightningOptimizer,
        optimizer_closure: Callable[[], Any] | None = None,
    ) -> None:
        """Override optimizer step to update the teacher parameters."""
        super().optimizer_step(
            epoch,
            batch_idx,
            optimizer,
            optimizer_closure=optimizer_closure,
        )
        self.update_teacher()



from .anomaly import AnomalyDataset
from .classification import ClassificationDataset, ImageClassificationListDataset, MultilabelClassificationDataset
from .patch import PatchSklearnClassificationTrainDataset
from .segmentation import SegmentationDataset, SegmentationDatasetMulticlass
from .ssl import TwoAugmentationDataset, TwoSetAugmentationDataset

__all__ = [
    "ImageClassificationListDataset",
    "ClassificationDataset",
    "SegmentationDataset",
    "SegmentationDatasetMulticlass",
    "PatchSklearnClassificationTrainDataset",
    "MultilabelClassificationDataset",
    "AnomalyDataset",
    "TwoAugmentationDataset",
    "TwoSetAugmentationDataset",
]



from typing import Any

import torch
from torch import nn


class SingleInputModel(nn.Module):
    """Model taking a single input."""

    def forward(self, x: Any):
        return x


class DoubleInputModel(nn.Module):
    """Model taking two inputs."""

    def forward(self, x: Any, y: Any):
        return x, y


class UnsupportedInputModel(nn.Module):
    """Model taking an unsupported input."""

    def forward(self, x: torch.Tensor, y: str):
        y = f"unsupported input: {y}"

        return x
